# pyeconomics/models/monetary_policy/monetary_policy_rules.py

from datetime import datetime

import matplotlib.pyplot as plt
import pandas as pd

from .balanced_approach_rule import (
    balanced_approach_rule, historical_balanced_approach_rule)
from .taylor_rule import taylor_rule, historical_taylor_rule
from .first_difference_rule import (
    first_difference_rule, historical_first_difference_rule)
from ...api import fetch_historical_fed_funds_rate, fred_client
from ...data.economic_indicators import EconomicIndicators
from ...data.model_parameters import (
    BalancedApproachRuleParameters,
    FirstDifferenceRuleParameters,
    TaylorRuleParameters,
)


def print_fred_series_names(
        inflation_series_id: str = 'PCETRIM12M159SFRBDAL',
        unemployment_rate_series_id: str = 'UNRATE',
        natural_unemployment_series_id: str = 'NROU',
        real_interest_rate_series_id: str = 'DFII10'
) -> None:
    """
    Print the FRED series IDs and their corresponding names for various
    economic indicators.

    Args:
        inflation_series_id (str): FRED Series ID for inflation data.
        unemployment_rate_series_id (str): FRED Series ID for unemployment rate.
        natural_unemployment_series_id (str): FRED Series ID for natural
            unemployment rate.
        real_interest_rate_series_id (str): FRED Series ID for long-term real
            interest rate.

    Returns:
        None
    """
    # Print the series names and their IDs
    print(
        f"Inflation Series ID:               "
        f"{fred_client.get_series_name(inflation_series_id)}")
    print(
        f"Unemployment Rate Series ID:       "
        f"{fred_client.get_series_name(unemployment_rate_series_id)}")
    print(
        f"Natural Unemployment Series ID:    "
        f"{fred_client.get_series_name(natural_unemployment_series_id)}")
    print(
        f"Real Interest Rate Series ID:      "
        f"{fred_client.get_series_name(real_interest_rate_series_id)}")


def print_verbose_output(
        estimates: pd.DataFrame,
        current_fed_rate: float,
        adjusted: bool = False
) -> None:
    """
    Format and print the verbose output of the interest rate policy estimates.

    Args:
        estimates (pd.DataFrame): DataFrame containing the policy estimates.
        current_fed_rate (float): The current Federal Funds Rate.
        adjusted (bool): Whether the output is for adjusted estimates.

    Returns:
        None
    """
    # Formatting and printing the verbose output
    as_of_date = datetime.now().strftime("%B %d, %Y")
    width = 85  # Total width of the box
    title = " Adjusted Interest Rate Policy Estimates " \
        if adjusted else " Interest Rate Policy Estimates "
    prescription_title = " Adjusted Policy Prescription " \
        if adjusted else " Policy Prescription "

    print("")
    print("┌" + "─" * (width - 2) + "┐")
    print("│" + title.center(width - 2) + "│")
    print("├" + "─" * (width - 2) + "┤")
    for rule, row in estimates.iterrows():
        line = f" {rule:69} {row['Estimate (%)']:.2f}%"
        print(f"│{line:<{width - 2}}│")
    print("├" + "─" * (width - 2) + "┤")
    ffr_description = "Federal Funds Rate (FFR)"
    ffr_value = f"{current_fed_rate:.2f}%"
    line = f" {ffr_description:<62} {ffr_value:>12}"
    print(f"│{line:<{width - 2}}│")
    print("├" + "─" * (width - 2) + "┤")
    line = f" As of Date {as_of_date:>64}"
    print(f"│{line:<{width - 2}}│")
    print("├" + "─" * (width - 2) + "┤")
    print("│" + prescription_title.center(width - 2) + "│")
    print("├" + "─" * (width - 2) + "┤")

    # Calculate the difference between each estimate and the current Fed rate
    for rule, row in estimates.iterrows():
        rate_difference = row['Estimate (%)'] - current_fed_rate
        rounded_difference = round(rate_difference * 4) / 4

        if rounded_difference > 0.125:
            suggestion = (f"suggests raising the rate by "
                          f"{rounded_difference:.2f}%.")
        elif rounded_difference < -0.125:
            suggestion = (f"suggests lowering the rate by "
                          f"{abs(rounded_difference):.2f}%.")
        else:
            suggestion = "suggests maintaining the current rate."

        # Combine rule and suggestion into one sentence and pad to fit the width
        full_suggestion = f" {rule} {suggestion}"
        print(f"│{full_suggestion:<{width - 2}}│")

    print("└" + "─" * (width - 2) + "┘")
    print("")


def calculate_policy_rule_estimates(
        indicators: EconomicIndicators = EconomicIndicators(),
        inflation_target: float = 2.0,
        rho: float = 0.0,
        elb: float = 0.125,
        apply_elb: bool = False,
        verbose: bool = False
) -> pd.DataFrame:
    """
    Calculate and return the monetary policy rule estimates as a DataFrame.

    Args:
        indicators (EconomicIndicators): Instance containing economic
            indicators.
        inflation_target (float): Target inflation rate.
        rho (float): Policy inertia coefficient. Defaults to 0.0 which means
            no inertia and no adjustment is made to Taylor Rule estimate. A
            value of 1.0 would mean full inertia and immediate central bank
            adjustment to the Taylor Rule estimate.
        elb (float): Effective lower bound for interest rates.
        apply_elb (bool): Whether to apply the effective lower bound
            constraint to the Taylor Rule estimate.
        verbose (bool): Whether to print verbose output.

    Returns:
        pd.DataFrame: DataFrame containing the policy estimates.
    """
    # Fetch the current Federal Funds Rate if not provided
    if indicators.current_fed_rate is None:
        indicators.current_fed_rate = fred_client.get_latest_value('DFEDTARU')

    tr_params = TaylorRuleParameters(
        inflation_target=inflation_target,
        rho=rho,
        elb=elb,
        apply_elb=apply_elb
    )

    bar_params = BalancedApproachRuleParameters(
        inflation_target=inflation_target,
        rho=rho,
        elb=elb,
        apply_elb=apply_elb
    )

    fdr_params = FirstDifferenceRuleParameters(
        inflation_target=inflation_target,
        rho=rho,
        elb=elb,
        apply_elb=apply_elb
    )

    # Current Taylor Rule calculation using FRED data
    tr_estimate = taylor_rule(indicators, tr_params)

    # Current Balanced Approach Rule calculation using FRED data
    bar_estimate = balanced_approach_rule(indicators, bar_params)

    # Current Balanced Approach (Shortfalls) Rule calculation using FRED data
    basr_params = bar_params
    basr_params.use_shortfalls_rule = True
    basr_estimate = balanced_approach_rule(
        indicators, bar_params)

    # Current First Difference Rule calculation using FRED data
    fdr_estimate = first_difference_rule(indicators, fdr_params)

    # Compile the estimates into a DataFrame
    estimates = pd.DataFrame(
        data=[
            tr_estimate,
            bar_estimate,
            basr_estimate,
            fdr_estimate
        ],
        columns=['Estimate (%)'],
        index=[
            'Taylor Rule (TR)',
            'Balanced Approach Rule (BAR)',
            'Balanced Approach Shortfalls Rule (BASR)',
            'First Difference Rule (FDR)'
        ]
    )

    if verbose and (rho > 0.0 or apply_elb):
        print_verbose_output(
            estimates, indicators.current_fed_rate, adjusted=True)
    elif verbose:
        print_verbose_output(estimates, indicators.current_fed_rate)
    return estimates


def calculate_historical_policy_rates(
        indicators: EconomicIndicators = EconomicIndicators(),
        inflation_target: float = 2.0,
        rho: float = 0.0,
        elb: float = 0.125,
        apply_elb: bool = False
) -> pd.DataFrame:
    """
    Calculate and return the historical monetary policy rule estimates as a
    DataFrame.

    Args:
        indicators (EconomicIndicators): Instance containing economic
            indicators.
        inflation_target (float): Target inflation rate.
        rho (float): Policy inertia coefficient. Defaults to 0.0 which means
            no inertia and no adjustment is made to Taylor Rule estimate. A
            value of 1.0 would mean full inertia and immediate central bank
            adjustment to the Taylor Rule estimate.
        elb (float): Effective lower bound for interest rates.
        apply_elb (bool): Whether to apply the effective lower bound
            constraint to the Taylor Rule estimate.

    Returns:
        pd.DataFrame: DataFrame containing the historical policy estimates.
    """
    tr_params = TaylorRuleParameters(
        inflation_target=inflation_target,
        rho=rho,
        elb=elb,
        apply_elb=apply_elb
    )

    bar_params = BalancedApproachRuleParameters(
        inflation_target=inflation_target,
        rho=rho,
        elb=elb,
        apply_elb=apply_elb
    )

    fdr_params = FirstDifferenceRuleParameters(
        inflation_target=inflation_target,
        rho=rho,
        elb=elb,
        apply_elb=apply_elb
    )

    # Historical Taylor Rule calculation using FRED data
    historical_tr = historical_taylor_rule(indicators, tr_params)

    # Historical BAR calculation using FRED data
    historical_bar = historical_balanced_approach_rule(indicators, bar_params)

    # Historical BASR calculation using FRED data
    basr_params = bar_params
    basr_params.use_shortfalls_rule = True
    historical_basr = historical_balanced_approach_rule(
        indicators, basr_params)

    # Historical First Difference Rule (FDR) calculation using FRED data
    historical_fdr = historical_first_difference_rule(indicators, fdr_params)

    # Combine historical rates into a single DataFrame
    historical_policy_rates = pd.concat([
        historical_tr['TaylorRule'],
        historical_tr['AdjustedTaylorRule'],
        historical_bar['BalancedApproachRule'],
        historical_bar['AdjustedBalancedApproachRule'],
        historical_basr['BalancedApproachShortfallsRule'],
        historical_basr['AdjustedBalancedApproachShortfallsRule'],
        historical_fdr['FirstDifferenceRule'],
        historical_fdr['AdjustedFirstDifferenceRule'],
        fetch_historical_fed_funds_rate()
    ], axis=1)

    return historical_policy_rates


def plot_historical_rule_estimates(
        historical_policy_rates: pd.DataFrame,
        adjusted: bool = False
) -> None:
    """
    Extract the time range from the data, plot the (adjusted or unadjusted)
    policy rules and the Federal Funds Rate, and return the date range.

    Args:
        historical_policy_rates (pd.DataFrame): DataFrame containing the
            historical policy rates.
        adjusted (bool): If True, plot the adjusted policy rates. If False,
            plot the unadjusted policy rates.

    Returns:
        None
    """
    # Determine which columns to use based on the adjusted parameter
    if adjusted:
        columns = [
            'AdjustedTaylorRule',
            'AdjustedBalancedApproachRule',
            'AdjustedBalancedApproachShortfallsRule',
            'AdjustedFirstDifferenceRule',
            'FedRate'
        ]
    else:
        columns = [
            'TaylorRule',
            'BalancedApproachRule',
            'BalancedApproachShortfallsRule',
            'FirstDifferenceRule',
            'FedRate'
        ]

    # Extracting the time range from the data
    start_date = historical_policy_rates.dropna().index.min()
    end_date = historical_policy_rates.dropna().index.max()
    date_range = (f"{start_date.strftime('%B %d, %Y')} to "
                  f"{end_date.strftime('%B %d, %Y')}")

    # Plotting Policy Rules and the Federal Funds Rate
    historical_policy_rates[columns].dropna().plot(
        figsize=(10, 5),  # Specifies the figure size
        grid=True  # Enables grid lines for better readability
    )

    title_prefix = 'Adjusted' if adjusted else 'Monetary'
    plt.title(
        f'{title_prefix} Policy Rule Estimates and Federal Funds Rate\n'
        f'{date_range}')
    plt.xlabel('Year')
    plt.ylabel('Interest Rate (%)')
    plt.legend([
        'Adjusted Taylor Rule' if adjusted
        else 'Taylor Rule',
        'Adjusted Balanced Approach Rule' if adjusted
        else 'Balanced Approach Rule',
        'Adjusted Balanced Approach (Shortfalls) Rule' if adjusted
        else 'Balanced Approach (Shortfalls) Rule',
        'Adjusted First Difference Rule' if adjusted
        else 'First Difference Rule',
        'Federal Funds Rate'
    ])

    # Adding the citation as a footnote
    plt.figtext(
        x=0.25,
        y=-0.01,
        s="Data Source: Federal Reserve Economic Data (FRED)",
        ha="center"
    )

    plt.show()  # Display the plot



# pyeconomics/utils/fdr_utils.py

from datetime import datetime


def verbose_first_difference_rule(data: dict):
    """
    Print verbose output for the First Difference Rule calculation.

    Args:
        data (dict): Dictionary containing the following keys:
            - current_inflation_rate (float): Current inflation rate.
            - inflation_target (float): Target inflation rate.
            - current_unemployment_rate (float): Current unemployment rate.
            - lagged_unemployment_rate (float): Lagged unemployment rate.
            - natural_unemployment_rate (float): Natural unemployment rate.
            - lagged_natural_unemployment_rate (float): Lagged natural
              unemployment rate.
            - current_fed_rate (float): Current Federal Reserve rate.
            - inflation_gap (float): Inflation gap.
            - current_unemployment_gap (float): Current unemployment gap.
            - lagged_unemployment_gap (float): Lagged unemployment gap.
            - alpha (float): Coefficient for the inflation gap.
            - rho (float): Policy inertia coefficient.
            - elb (float): Effective Lower Bound (ELB) rate.
            - apply_elb (bool): Whether to apply the ELB adjustment.
            - unadjusted_fdr_rule (float): Unadjusted First Difference Rule
              estimate.
            - adjusted_fdr_rule_after_elb (float): First Difference Rule
              adjusted for the Effective Lower Bound (ELB).
            - adjusted_fdr_rule_after_inertia (float): First Difference Rule
              adjusted for policy inertia.

    Returns:
        None
    """
    current_date = datetime.now().strftime("%B %d, %Y")
    print(
        "\n==== Economic Indicators ==========================================="
        "=")
    print("Current Inflation:                               {:.2f}%".format(
        data['current_inflation_rate']))
    print("Target Inflation:                                {:.2f}%".format(
        data['inflation_target']))
    print("Current Unemployment Rate:                       {:.2f}%".format(
        data['current_unemployment_rate']))
    print("Lagged Unemployment Rate:                        {:.2f}%".format(
        data['lagged_unemployment_rate']))
    print("Natural Unemployment Rate:                       {:.2f}%".format(
        data['natural_unemployment_rate']))
    print("Lagged Natural Unemployment Rate:                {:.2f}%".format(
        data['lagged_natural_unemployment_rate']))
    print("Last Fed Rate:                                   {:.2f}%".format(
        data['current_fed_rate']))
    print("As of Date:                                      {}".format(
        current_date))

    print(
        "\n==== Gaps =========================================================="
        "=")
    print("Inflation Gap:                                   {:.2f}%".format(
        data['inflation_gap']))
    print("Current Unemployment Gap:                        {:.2f}%".format(
        data['current_unemployment_gap']))
    print("Lagged Unemployment Gap:                         {:.2f}%".format(
        data['lagged_unemployment_gap']))

    print(
        "\n==== First Difference Rule (FDR) ==================================="
        "=")
    print("  Last Fed Rate:                                 {:.2f}%".format(
        data['current_fed_rate']))
    print("  Alpha * Inflation Gap:                         "
          "+ {:.2f} * {:.2f}%".format(
            data['alpha'], data['inflation_gap']))
    print("  Current Unemployment Gap Adjustment:           + {:.2f}%".format(
        data['current_unemployment_gap']))
    print("  Last Year Unemployment Gap Adjustment:         - {:.2f}%".format(
        data['lagged_unemployment_gap']))
    print(
        "---------------------------------------------------------------------")
    print("  Unadjusted FDR Estimate:                       {:.2f}%".format(
        data['unadjusted_fdr_rule']))

    print(
        "\n==== Adjusted First Difference Rule ================================"
        "=")
    if data['apply_elb']:
        print("  Effective Lower Bound (ELB) Adjustment:")
        print(
            "  Maximum of FDR or ELB:                         "
            "max({:.2f}%, {:.2f}%)".format(
                data['unadjusted_fdr_rule'], data['elb']))
        print(
            "------------------------------------------------------------------"
            "---")
        print("  FDR Adjusted for ELB:                          {:.2f}%".format(
            data['adjusted_fdr_rule_after_elb']))

    print("\n  Policy Inertia Adjustment:")
    print("  Policy Inertia Coefficient (rho):              {:.2f}".format(
        data['rho']))
    print("  Last Fed Rate:                                 * {:.2f}%".format(
        data['current_fed_rate']))
    print("  Adjustment Coefficient (1 - rho):              "
          "+ (1 - {:.2f})".format(data['rho']))
    print(f"  FDR Adjusted for ELB:                         "
          f" * {data['adjusted_fdr_rule_after_elb']:.2f}%")
    print(
        "---------------------------------------------------------------------")
    print("  Adjusted FDR Estimate:                         {:.2f}%".format(
        data['adjusted_fdr_rule_after_inertia']))

    # Policy Prescription section
    print("\n==== Policy Prescription ========================================="
          "========")
    rate_difference = (data['adjusted_fdr_rule_after_inertia'] -
                       data['current_fed_rate'])
    rounded_difference = round(rate_difference * 4) / 4

    if rounded_difference > 0.125:
        print(f"  The Adjusted FDR Estimate is "
              f"{rate_difference:.2f}% higher than the Current Fed \n"
              f"  Rate. The Fed should consider raising the interest "
              f"rate by {rounded_difference:.2f}%.")
    elif rounded_difference < -0.125:
        print(f"  The Adjusted FDR Estimate is "
              f"{abs(rate_difference):.2f}% lower than the Current Fed \n"
              f"  Rate. The Fed should consider lowering the interest "
              f"rate by {abs(rounded_difference):.2f}%.")
    else:
        print(f"  The Adjusted FDR Estimate is equal to the Current "
              f"Fed Rate.\n  The Fed should maintain the current interest "
              f"rate.")



# -- Path setup --------------------------------------------------------------

import os
import re
import sys


# Function to extract the version from __version__.py
def get_version():
    here = os.path.abspath(os.path.dirname(__file__))
    with open(
            os.path.join(here, '../__version__.py'),
            encoding='utf-8') as f:
        version_file_contents = f.read()
    # Use regular expression to extract version string
    version_match = re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]",
        version_file_contents,
        re.M
    )
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


sys.path.insert(0, os.path.abspath('../..'))

# -- Project information -----------------------------------------------------

project = 'PyEconomics'
author = 'Nathan Ramos, CFA'

# The full version, including alpha/beta/rc tags
version = get_version()
release = get_version()

# Avoid shadowing the built-in 'copyright' name
project_copyright = '2024, Nathan Ramos, CFA'

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx_rtd_theme'
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# Exclude members from being documented
autodoc_default_options = {
    'exclude-members': 'API_BASE_URL, PRO_API_BASE_URL'
}

# Configure the source suffix to include both 'data.rst' and '.md' file types
source_suffix = ['.rst', '.md']
source_encoding = 'utf-8'
master_doc = 'index'

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']
html_logo = '_static/logo.png'

# -- Options for EPUB output -------------------------------------------------

# EPUB settings
epub_title = project
epub_author = author
epub_publisher = author
epub_copyright = '2024, Nathan Ramos, CFA'



# pyeconomics/utils/__init__.py

from .bar_utils import verbose_balanced_approach_rule
from .fdr_utils import verbose_first_difference_rule
from .tr_utils import verbose_taylor_rule

__all__ = ['verbose_balanced_approach_rule', 'verbose_first_difference_rule',
           'verbose_taylor_rule']



# pyeconomics/models/monetary_policy/__init__.py

from .balanced_approach_rule import (
    balanced_approach_rule, historical_balanced_approach_rule,
    plot_historical_bar_basr_rule
)

from .first_difference_rule import (
    first_difference_rule, historical_first_difference_rule,
    plot_historical_fdr
)

from .taylor_rule import (
    taylor_rule, historical_taylor_rule, plot_historical_taylor_rule
)

from .monetary_policy_rules import (
    print_fred_series_names, print_verbose_output,
    calculate_policy_rule_estimates, calculate_historical_policy_rates,
    plot_historical_rule_estimates
)

__all__ = [
    'balanced_approach_rule',
    'calculate_historical_policy_rates',
    'calculate_policy_rule_estimates',
    'first_difference_rule',
    'historical_balanced_approach_rule',
    'historical_first_difference_rule',
    'historical_taylor_rule',
    'plot_historical_rule_estimates',
    'plot_historical_bar_basr_rule',
    'plot_historical_fdr',
    'plot_historical_taylor_rule',
    'print_fred_series_names',
    'print_verbose_output',
    'taylor_rule',
]



# test/conftest.py

import pytest
from unittest.mock import patch, MagicMock


@pytest.fixture(autouse=True)
def mock_keyring():
    with patch('keyring.get_password') as mock_get_password:
        mock_get_password.return_value = 'dummy_fred_api_key'
        yield


@pytest.fixture(autouse=True)
def mock_fred_client():
    with patch('pyeconomics.api.fred_api.FredClient.__new__') as mock_new:
        mock_instance = MagicMock()
        mock_instance.client = MagicMock()
        mock_new.return_value = mock_instance
        yield mock_instance



# pyeconomics/api/cache_manager.py

import os
import pickle
from datetime import datetime, timedelta
from hashlib import sha256
from typing import Any

# Define the cache directory relative to the root of the project
CACHE_DIR = os.path.join(
    os.path.dirname(os.path.dirname(os.path.dirname(__file__))), 'cache')

if not os.path.exists(CACHE_DIR):
    os.makedirs(CACHE_DIR)


def cache_filename(key: str) -> str:
    """Generate a filename for the cache based on a hashed key.

    Args:
        key (str): The key to hash for generating the cache filename.

    Returns:
        str: The path to the cache file.
    """
    hashed_key = sha256(key.encode('utf-8')).hexdigest()
    return os.path.join(CACHE_DIR, f"{hashed_key}.pkl")


def save_to_cache(key: str, data: Any) -> None:
    """Save data to the cache.

    Args:
        key (str): The key for the cache entry.
        data: The data to be cached.

    Returns:
        None
    """
    filename = cache_filename(key)
    with open(filename, 'wb') as f:
        pickle.dump(data, f)


def load_from_cache(key: str, expiry: timedelta = timedelta(days=1)):
    """Load data from the cache if available and not expired.

    Args:
        key (str): The key for the cache entry.
        expiry (timedelta): The expiration time for the cache entry.

    Returns:
        The cached data if available and not expired, otherwise None.
    """
    filename = cache_filename(key)
    if os.path.exists(filename):
        file_mtime = datetime.fromtimestamp(os.path.getmtime(filename))
        if datetime.now() - file_mtime < expiry:
            with open(filename, 'rb') as f:
                return pickle.load(f)
    return None



# pyeconomics/data/__init__.py

from .economic_indicators import EconomicIndicators
from .model_parameters import TaylorRuleParameters
from .model_parameters import BalancedApproachRuleParameters
from .model_parameters import FirstDifferenceRuleParameters

__all__ = [
    'EconomicIndicators',
    'TaylorRuleParameters',
    'BalancedApproachRuleParameters',
    'FirstDifferenceRuleParameters'
]



# pyeconomics/models/monetary_policy/balanced_approach_rule.py

import matplotlib.pyplot as plt
import pandas as pd
from typing import Optional

from pyeconomics.api import fetch_historical_fed_funds_rate, fred_client
from pyeconomics.data.economic_indicators import EconomicIndicators
from pyeconomics.data.model_parameters import BalancedApproachRuleParameters
from pyeconomics.utils import verbose_balanced_approach_rule


def balanced_approach_rule(
        indicators: EconomicIndicators = EconomicIndicators(),
        params: BalancedApproachRuleParameters =
        BalancedApproachRuleParameters(),
        verbose: Optional[bool] = None
) -> float:
    """
    Computes the Balanced Approach Rule interest rate based on economic
    indicators.

    Args:
        indicators (EconomicIndicators): Economic indicators data class.
        params (BalancedApproachRuleParameters): Balanced Approach Rule
            parameters data class.
        verbose (bool, optional): Whether to print verbose output. If not
            provided, defaults to the value in params. Defaults to None.

    Returns:
        float: Balanced Approach Rule interest rate estimate.
    """
    # Override params.verbose if verbose is explicitly provided
    verbose = verbose if verbose is not None else params.verbose

    # Fetch data if not provided
    indicators.current_inflation_rate = fred_client.get_data_or_fetch(
        indicators.current_inflation_rate,
        indicators.inflation_series_id)
    indicators.current_unemployment_rate = fred_client.get_data_or_fetch(
        indicators.current_unemployment_rate,
        indicators.unemployment_rate_series_id)
    indicators.natural_unemployment_rate = fred_client.get_data_or_fetch(
        indicators.natural_unemployment_rate,
        indicators.natural_unemployment_series_id)
    indicators.long_term_real_interest_rate = fred_client.get_data_or_fetch(
        indicators.long_term_real_interest_rate,
        indicators.real_interest_rate_series_id)

    if indicators.current_fed_rate is None:
        indicators.current_fed_rate = fred_client.get_latest_value('DFEDTARU')

    if None in (indicators.current_inflation_rate,
                indicators.current_unemployment_rate,
                indicators.natural_unemployment_rate,
                indicators.long_term_real_interest_rate,
                indicators.current_fed_rate):
        raise ValueError("Required economic data is missing.")

    # Calculate gaps and Balanced Approach Rule estimate
    inflation_gap = indicators.current_inflation_rate - params.inflation_target
    unemployment_gap = (indicators.natural_unemployment_rate -
                        indicators.current_unemployment_rate)

    # Apply shortfalls rule
    if params.use_shortfalls_rule:
        unemployment_gap = min(0.0, unemployment_gap)

    unadjusted_rate = (indicators.long_term_real_interest_rate +
                       indicators.current_inflation_rate +
                       params.alpha * inflation_gap +
                       params.beta * unemployment_gap)

    # Apply effective lower bound constraint
    if params.apply_elb:
        adjusted_rate_after_elb = max(unadjusted_rate, params.elb)
    else:
        adjusted_rate_after_elb = unadjusted_rate

    # Apply policy inertia
    adjusted_rate_after_inertia = (
        params.rho * indicators.current_fed_rate +
        (1 - params.rho) * adjusted_rate_after_elb
    )

    # Verbose output
    if verbose:
        data = {
            'current_inflation_rate': indicators.current_inflation_rate,
            'inflation_target': params.inflation_target,
            'current_unemployment_rate': indicators.current_unemployment_rate,
            'natural_unemployment_rate': indicators.natural_unemployment_rate,
            'long_term_real_interest_rate':
                indicators.long_term_real_interest_rate,
            'current_fed_rate': indicators.current_fed_rate,
            'inflation_gap': inflation_gap,
            'unemployment_gap': unemployment_gap,
            'unadjusted_rate': unadjusted_rate,
            'adjusted_rate_after_elb': adjusted_rate_after_elb,
            'adjusted_rate_after_inertia': adjusted_rate_after_inertia,
            'rho': params.rho,
            'alpha': params.alpha,
            'beta': params.beta,
            'elb': params.elb,
            'apply_elb': params.apply_elb,
            'use_shortfalls_rule': params.use_shortfalls_rule
        }
        verbose_balanced_approach_rule(data)

    return round(adjusted_rate_after_inertia, 2)


def historical_balanced_approach_rule(
        indicators: EconomicIndicators,
        params: BalancedApproachRuleParameters
) -> pd.DataFrame:
    """
    Computes historical Balanced Approach Rule interest rates using economic
    indicators up to the last date with available real interest rate data.

    Args:
        indicators (EconomicIndicators): Economic indicators data class.
        params (BalancedApproachRuleParameters): Balanced Approach Rule
            parameters data class.

    Returns:
        pd.DataFrame: DataFrame with computed Balanced Approach Rule rates.
    """
    # Fetch historical data for all series
    inflation = fred_client.fetch_data(indicators.inflation_series_id)
    unemployment_rate = fred_client.fetch_data(
        indicators.unemployment_rate_series_id)
    natural_unemployment = fred_client.fetch_data(
        indicators.natural_unemployment_series_id)
    real_interest_rate = fred_client.fetch_data(
        indicators.real_interest_rate_series_id)
    fed_rate = fetch_historical_fed_funds_rate()

    # Combine into a DataFrame
    data = pd.DataFrame({
        'Inflation': inflation,
        'UnemploymentRate': unemployment_rate,
        'NaturalUnemploymentRate': natural_unemployment,
        'RealInterestRate': real_interest_rate,
        'FedRate': fed_rate
    })

    # Determine the cutoff date based on RealInterestRate data
    last_real_interest_date = data['RealInterestRate'].last_valid_index()
    data = data.loc[:last_real_interest_date]

    # Handle missing data
    data.ffill(inplace=True)
    data.dropna(inplace=True)

    # Calculate gaps and Balanced Approach Rule estimation
    data['InflationGap'] = data['Inflation'] - params.inflation_target
    data['UnemploymentGap'] = (data['NaturalUnemploymentRate'] -
                               data['UnemploymentRate'])

    # Apply shortfalls rule
    if params.use_shortfalls_rule:
        data['UnemploymentGap'] = data['UnemploymentGap'].apply(
            lambda x: min(0, x))
        rule_name = 'BalancedApproachShortfallsRule'
    else:
        rule_name = 'BalancedApproachRule'

    data[rule_name] = (data['RealInterestRate'] +
                       data['Inflation'] +
                       params.alpha * data['InflationGap'] +
                       params.beta * data['UnemploymentGap'])

    adjusted_rule_name = 'Adjusted' + rule_name

    # Apply effective lower bound
    if params.apply_elb:
        data[adjusted_rule_name] = data[rule_name].apply(
            lambda x: max(x, params.elb))
    else:
        data[adjusted_rule_name] = data[rule_name]

    # Apply policy inertia
    data[adjusted_rule_name] = (
            params.rho * data['FedRate'] +
            (1 - params.rho) * data[adjusted_rule_name])

    return data.round(2)


def plot_historical_bar_basr_rule(
    historical_rates: pd.DataFrame,
    adjusted: bool = False
) -> None:
    """
    Extract the time range from the data, plot either the Balanced Approach
    Rule (BAR) and Balanced Approach Shortfalls Rule (BASR) estimates or their
    adjusted versions along with the Federal Funds Rate, and display the date
    range in the plot title.

    Args:
        historical_rates (pd.DataFrame): DataFrame containing the historical
            rates including BAR, BASR estimates, and the Federal Funds Rate,
            along with their adjusted versions if applicable.
        adjusted (bool): A flag to determine whether to plot the adjusted or
            unadjusted rates.

    Returns:
        None
    """
    if adjusted:
        columns = [
            'AdjustedBalancedApproachRule',
            'AdjustedBalancedApproachShortfallsRule',
            'FedRate'
        ]
        title_prefix = "Adjusted "
    else:
        columns = [
            'BalancedApproachRule',
            'BalancedApproachShortfallsRule',
            'FedRate'
        ]
        title_prefix = ""

    # Select relevant data
    plot_data = historical_rates[columns].copy()

    # Extracting the time range from the data
    start_date = plot_data.dropna().index.min()
    end_date = plot_data.dropna().index.max()
    date_range = (f"{start_date.strftime('%B %d, %Y')} to "
                  f"{end_date.strftime('%B %d, %Y')}")

    # Plotting BAR and BASR Rule estimates and the Federal Funds Rate
    plot_data.dropna().plot(
        figsize=(10, 5),  # Specifies the figure size
        grid=True  # Enables grid lines for better readability
    )

    plt.title(f'{title_prefix}BAR and BASR Rule Estimates and Federal Funds '
              f'Rate\n{date_range}')
    plt.xlabel('Year')
    plt.ylabel('Interest Rate (%)')
    plt.legend([
        f'{title_prefix}BAR Rule',
        f'{title_prefix}BASR Rule',
        'Federal Funds Rate']
    )

    # Adding the citation as a footnote
    plt.figtext(
        x=0.25,
        y=-0.01,
        s="Data Source: Federal Reserve Economic Data (FRED)",
        ha="center"
    )

    plt.show()  # Display the plot



# pyeconomics/models/monetary_policy/first_difference_rule.py

import logging
import matplotlib.pyplot as plt
import pandas as pd
from typing import Optional

from pyeconomics.api import fetch_historical_fed_funds_rate, fred_client
from pyeconomics.data.economic_indicators import EconomicIndicators
from pyeconomics.data.model_parameters import FirstDifferenceRuleParameters
from pyeconomics.utils import verbose_first_difference_rule


def first_difference_rule(
        indicators: EconomicIndicators = EconomicIndicators(),
        params: FirstDifferenceRuleParameters = FirstDifferenceRuleParameters(),
        verbose: Optional[bool] = None
) -> float:
    """
    Computes the First-Difference Rule interest rate based on economic
    indicators.

    Args:
        indicators (EconomicIndicators): Economic indicators data class.
        params (FirstDifferenceRuleParameters): First Difference Rule
            parameters data class.
        verbose (bool, optional): Whether to print verbose output. If not
            provided, defaults to the value in params. Defaults to None.

    Returns:
        float: First-Difference Rule interest rate estimate.
    """
    try:
        # Override params.verbose if verbose is explicitly provided
        verbose = verbose if verbose is not None else params.verbose

        # Fetch data if not provided
        indicators.current_inflation_rate = fred_client.get_data_or_fetch(
            indicators.current_inflation_rate, indicators.inflation_series_id)
        indicators.current_unemployment_rate = fred_client.get_data_or_fetch(
            indicators.current_unemployment_rate,
            indicators.unemployment_rate_series_id)
        indicators.natural_unemployment_rate = fred_client.get_data_or_fetch(
            indicators.natural_unemployment_rate,
            indicators.natural_unemployment_series_id)
        indicators.current_fed_rate = fred_client.get_data_or_fetch(
            indicators.current_fed_rate, 'DFEDTARU')

        indicators.lagged_unemployment_rate = fred_client.get_data_or_fetch(
            indicators.lagged_unemployment_rate,
            indicators.unemployment_rate_series_id, periods=-12)
        indicators.lagged_natural_unemployment_rate = (
            fred_client.get_data_or_fetch(
                indicators.lagged_natural_unemployment_rate,
                indicators.natural_unemployment_series_id, periods=-4)
        )

    except Exception as e:
        logging.error(f"Error fetching data: {e}")
        raise ValueError("Missing or invalid data")

    # Calculate components of the First-Difference Rule
    inflation_gap = indicators.current_inflation_rate - params.inflation_target
    current_unemployment_gap = (indicators.natural_unemployment_rate -
                                indicators.current_unemployment_rate)
    lagged_unemployment_gap = (indicators.lagged_natural_unemployment_rate -
                               indicators.lagged_unemployment_rate)

    unadjusted_fdr_rule = (indicators.current_fed_rate +
                           params.alpha * inflation_gap +
                           current_unemployment_gap -
                           lagged_unemployment_gap)

    # Apply an effective lower bound (ELB)
    if params.apply_elb:
        adjusted_fdr_rule_after_elb = max(unadjusted_fdr_rule, params.elb)
    else:
        adjusted_fdr_rule_after_elb = unadjusted_fdr_rule

    # Apply policy inertia
    adjusted_fdr_rule_after_inertia = (
        params.rho * indicators.current_fed_rate +
        (1 - params.rho) * adjusted_fdr_rule_after_elb
    )

    # Verbose output
    if verbose:
        data = {
            'current_inflation_rate': indicators.current_inflation_rate,
            'inflation_target': params.inflation_target,
            'current_unemployment_rate': indicators.current_unemployment_rate,
            'lagged_unemployment_rate': indicators.lagged_unemployment_rate,
            'natural_unemployment_rate': indicators.natural_unemployment_rate,
            'lagged_natural_unemployment_rate':
                indicators.lagged_natural_unemployment_rate,
            'current_fed_rate': indicators.current_fed_rate,
            'inflation_gap': inflation_gap,
            'current_unemployment_gap': current_unemployment_gap,
            'lagged_unemployment_gap': lagged_unemployment_gap,
            'unadjusted_fdr_rule': unadjusted_fdr_rule,
            'adjusted_fdr_rule_after_elb': adjusted_fdr_rule_after_elb,
            'adjusted_fdr_rule_after_inertia': adjusted_fdr_rule_after_inertia,
            'alpha': params.alpha,
            'rho': params.rho,
            'elb': params.elb,
            'apply_elb': params.apply_elb
        }
        verbose_first_difference_rule(data)

    return round(adjusted_fdr_rule_after_inertia, 2)


def historical_first_difference_rule(
        indicators: EconomicIndicators,
        params: FirstDifferenceRuleParameters
) -> pd.DataFrame:
    """
    Computes historical First Difference Rule interest rates using economic
    indicators up to the last date with available data.

    Args:
        indicators (EconomicIndicators): Economic indicators data class.
        params (FirstDifferenceRuleParameters): First Difference Rule
            parameters data class.

    Returns:
        pd.DataFrame: DataFrame with computed First Difference Rule rates.
    """
    try:
        # Fetch historical data for all series
        inflation = fred_client.fetch_data(indicators.inflation_series_id)
        unemployment_rate = fred_client.fetch_data(
            indicators.unemployment_rate_series_id)
        lagged_unemployment_rate = unemployment_rate.shift(12)
        natural_unemployment = fred_client.fetch_data(
            indicators.natural_unemployment_series_id)
        lagged_natural_unemployment = natural_unemployment.shift(4)
        fed_rate = fetch_historical_fed_funds_rate()

        # Check for missing data
        if (inflation is None or
                unemployment_rate is None or
                lagged_unemployment_rate is None or
                natural_unemployment is None or
                lagged_natural_unemployment is None or
                fed_rate is None):
            raise ValueError("Missing or invalid data")
    except Exception as e:
        logging.error(f"Error fetching historical data: {e}")
        raise ValueError("Missing or invalid data")

    # Combine into a DataFrame
    data = pd.DataFrame({
        'Inflation': inflation,
        'UnemploymentRate': unemployment_rate,
        'LaggedUnemploymentRate': lagged_unemployment_rate,
        'NaturalUnemploymentRate': natural_unemployment,
        'LaggedNaturalUnemploymentRate': lagged_natural_unemployment,
        'FedRate': fed_rate
    })

    # Compute lagged unemployment gap 12 months ago
    data['LaggedUnemploymentGap'] = (
        data['LaggedNaturalUnemploymentRate'] - data['LaggedUnemploymentRate'])

    # Determine the cutoff date based on FedRate data
    fed_rate_date = data['FedRate'].last_valid_index()
    data = data.loc[:fed_rate_date]

    # Handle missing data
    data.ffill(inplace=True)
    data.dropna(inplace=True)

    # Calculate historical gaps and First Difference Rule estimation
    data['InflationGap'] = data['Inflation'] - params.inflation_target
    data['UnemploymentGap'] = (
        data['NaturalUnemploymentRate'] - data['UnemploymentRate'])

    # First Difference Rule Calculation
    data['FirstDifferenceRule'] = (
        data['FedRate'] +
        params.alpha * data['InflationGap'] +
        data['UnemploymentGap'] -
        data['LaggedUnemploymentGap']
    )

    # Apply effective lower bound constraint
    if params.apply_elb:
        data['AdjustedFirstDifferenceRule'] = data['FirstDifferenceRule'].apply(
            lambda x: max(x, params.elb))
    else:
        data['AdjustedFirstDifferenceRule'] = data['FirstDifferenceRule']

    # Apply policy inertia
    data['AdjustedFirstDifferenceRule'] = (
        params.rho * data['FedRate'] +
        (1 - params.rho) * data['AdjustedFirstDifferenceRule'])

    return data.round(2)


def plot_historical_fdr(
        historical_rates: pd.DataFrame
) -> None:
    """
    Extracts the time range from the data and plots the First Difference
    Rule (FDR) estimates along with the Federal Funds Rate.

    Args:
        historical_rates (pd.DataFrame): DataFrame containing the
            historical rates including FDR estimates and the Federal
            Funds Rate.

    Returns:
        None
    """
    # Extracting the time range from the data
    start_date = historical_rates.dropna().index.min()
    end_date = historical_rates.dropna().index.max()
    date_range = (f"{start_date.strftime('%B %d, %Y')} to "
                  f"{end_date.strftime('%B %d, %Y')}")

    # Plotting First Difference Rule (FDR) estimates and the Federal Funds Rate
    historical_rates[['FirstDifferenceRule',
                      'AdjustedFirstDifferenceRule',
                      'FedRate']].dropna().plot(
        figsize=(10, 5),  # Specifies the figure size
        grid=True  # Enables grid lines for better readability
    )

    plt.title(f'First Difference Rule Estimates and Federal Funds Rate\n'
              f'{date_range}')

    plt.xlabel('Year')
    plt.ylabel('Interest Rate (%)')
    plt.legend(['First Difference Rule', 'Adjusted First Difference Rule',
                'Federal Funds Rate'])

    # Adding the citation as a footnote
    plt.figtext(
        x=0.25,
        y=-0.01,
        s="Data Source: Federal Reserve Economic Data (FRED)",
        ha="center"
    )

    plt.show()  # Display the plot



# pyeconomics/__init__.py

# API imports
from .api import fetch_historical_fed_funds_rate, FredClient, fred_client
from .api import load_from_cache, save_to_cache

# Data imports
from .data import BalancedApproachRuleParameters
from .data import EconomicIndicators
from .data import FirstDifferenceRuleParameters
from .data import TaylorRuleParameters

# Models imports
from .models.monetary_policy import balanced_approach_rule
from .models.monetary_policy import calculate_historical_policy_rates
from .models.monetary_policy import calculate_policy_rule_estimates
from .models.monetary_policy import first_difference_rule
from .models.monetary_policy import historical_balanced_approach_rule
from .models.monetary_policy import historical_first_difference_rule
from .models.monetary_policy import historical_taylor_rule
from .models.monetary_policy import plot_historical_rule_estimates
from .models.monetary_policy import plot_historical_bar_basr_rule
from .models.monetary_policy import plot_historical_fdr
from .models.monetary_policy import plot_historical_taylor_rule
from .models.monetary_policy import print_fred_series_names
from .models.monetary_policy import print_verbose_output
from .models.monetary_policy import taylor_rule

# Utilities imports
from .utils.bar_utils import verbose_balanced_approach_rule
from .utils.fdr_utils import verbose_first_difference_rule
from .utils.tr_utils import verbose_taylor_rule

# Exported symbols
__all__ = [
    'BalancedApproachRuleParameters',
    'balanced_approach_rule',
    'calculate_historical_policy_rates',
    'calculate_policy_rule_estimates',
    'EconomicIndicators',
    'fetch_historical_fed_funds_rate',
    'FirstDifferenceRuleParameters',
    'first_difference_rule',
    'fred_client',
    'FredClient',
    'historical_balanced_approach_rule',
    'historical_first_difference_rule',
    'historical_taylor_rule',
    'load_from_cache',
    'plot_historical_rule_estimates',
    'plot_historical_bar_basr_rule',
    'plot_historical_fdr',
    'plot_historical_taylor_rule',
    'print_fred_series_names',
    'print_verbose_output',
    'save_to_cache',
    'TaylorRuleParameters',
    'taylor_rule',
    'verbose_balanced_approach_rule',
    'verbose_first_difference_rule',
    'verbose_taylor_rule'
]



# pyeconomics/models/monetary_policy/taylor_rule.py

import matplotlib.pyplot as plt
import pandas as pd
from typing import Optional

from pyeconomics.api import fetch_historical_fed_funds_rate, fred_client
from pyeconomics.data.economic_indicators import EconomicIndicators
from pyeconomics.data.model_parameters import TaylorRuleParameters
from pyeconomics.utils import verbose_taylor_rule


def taylor_rule(
        indicators: EconomicIndicators = EconomicIndicators(),
        params: TaylorRuleParameters = TaylorRuleParameters(),
        verbose: Optional[bool] = None
) -> float:
    """
    Computes the Taylor Rule interest rate based on economic indicators.

    Args:
        indicators (EconomicIndicators): Economic indicators data class.
        params (TaylorRuleParameters): Taylor Rule parameters data class.
        verbose (bool, optional): Whether to print verbose output. If not
            provided, defaults to the value in params. Defaults to None.

    Returns:
        float: Taylor Rule interest rate estimate.
    """
    # Override params.verbose if verbose is explicitly provided
    verbose = verbose if verbose is not None else params.verbose

    # Fetch data if not provided
    indicators.current_inflation_rate = fred_client.get_data_or_fetch(
        indicators.current_inflation_rate,
        indicators.inflation_series_id
    )
    indicators.current_unemployment_rate = fred_client.get_data_or_fetch(
        indicators.current_unemployment_rate,
        indicators.unemployment_rate_series_id
    )
    indicators.natural_unemployment_rate = fred_client.get_data_or_fetch(
        indicators.natural_unemployment_rate,
        indicators.natural_unemployment_series_id
    )
    indicators.long_term_real_interest_rate = fred_client.get_data_or_fetch(
        indicators.long_term_real_interest_rate,
        indicators.real_interest_rate_series_id
    )

    if indicators.current_fed_rate is None:
        indicators.current_fed_rate = fred_client.get_latest_value('DFEDTARU')

    if None in (
        indicators.current_inflation_rate,
        indicators.current_unemployment_rate,
        indicators.natural_unemployment_rate,
        indicators.long_term_real_interest_rate,
        indicators.current_fed_rate
    ):
        raise ValueError("Required economic data is missing.")

    # Calculate gaps and Taylor Rule estimate
    inflation_gap = indicators.current_inflation_rate - params.inflation_target
    unemployment_gap = (indicators.natural_unemployment_rate -
                        indicators.current_unemployment_rate)

    unadjusted_taylor_rule = (
        indicators.long_term_real_interest_rate +
        indicators.current_inflation_rate +
        params.alpha * inflation_gap +
        params.beta * params.okun_factor * unemployment_gap
    )

    # Apply effective lower bound constraint
    if params.apply_elb:
        adjusted_taylor_rule_after_elb = max(unadjusted_taylor_rule, params.elb)
    else:
        adjusted_taylor_rule_after_elb = unadjusted_taylor_rule

    # Apply policy inertia
    adjusted_taylor_rule_after_inertia = (
        params.rho * indicators.current_fed_rate +
        (1 - params.rho) * adjusted_taylor_rule_after_elb
    )

    if verbose:
        data = {
            'current_inflation_rate': indicators.current_inflation_rate,
            'inflation_target': params.inflation_target,
            'current_unemployment_rate': indicators.current_unemployment_rate,
            'natural_unemployment_rate': indicators.natural_unemployment_rate,
            'long_term_real_interest_rate':
                indicators.long_term_real_interest_rate,
            'current_fed_rate': indicators.current_fed_rate,
            'inflation_gap': inflation_gap,
            'unemployment_gap': unemployment_gap,
            'unadjusted_taylor_rule': unadjusted_taylor_rule,
            'adjusted_taylor_rule_after_elb': adjusted_taylor_rule_after_elb,
            'adjusted_taylor_rule_after_inertia':
                adjusted_taylor_rule_after_inertia,
            'rho': params.rho,
            'alpha': params.alpha,
            'beta': params.beta,
            'elb': params.elb,
            'apply_elb': params.apply_elb,
            'okun_factor': params.okun_factor
        }
        verbose_taylor_rule(data)

    return round(adjusted_taylor_rule_after_inertia, 2)


def historical_taylor_rule(
        indicators: EconomicIndicators,
        params: TaylorRuleParameters
) -> pd.DataFrame:
    """
    Computes historical Taylor Rule interest rates using economic indicators
    up to the last date with available real interest rate data.

    Args:
        indicators (EconomicIndicators): Economic indicators data class.
        params (TaylorRuleParameters): Taylor Rule parameters data class.

    Returns:
        pd.DataFrame: DataFrame with computed Taylor Rule rates.
    """
    # Fetch historical data for all series
    inflation = fred_client.fetch_data(indicators.inflation_series_id)
    unemployment_rate = fred_client.fetch_data(
        indicators.unemployment_rate_series_id)
    natural_unemployment = fred_client.fetch_data(
        indicators.natural_unemployment_series_id)
    real_interest_rate = fred_client.fetch_data(
        indicators.real_interest_rate_series_id)
    fed_rate = fetch_historical_fed_funds_rate()

    # Combine into a DataFrame
    data = pd.DataFrame({
        'Inflation': inflation,
        'UnemploymentRate': unemployment_rate,
        'NaturalUnemploymentRate': natural_unemployment,
        'RealInterestRate': real_interest_rate,
        'FedRate': fed_rate
    })

    # Determine the cutoff date based on RealInterestRate data
    last_real_interest_date = data['RealInterestRate'].last_valid_index()
    data = data.loc[:last_real_interest_date]

    # Handle missing data
    data.ffill(inplace=True)
    data.dropna(inplace=True)

    # Calculate gaps and Taylor Rule estimation
    data['InflationGap'] = data['Inflation'] - params.inflation_target
    data['UnemploymentGap'] = (data['NaturalUnemploymentRate'] -
                               data['UnemploymentRate'])
    data['TaylorRule'] = (data['RealInterestRate'] +
                          data['Inflation'] +
                          params.alpha * data['InflationGap'] +
                          params.beta * params.okun_factor *
                          data['UnemploymentGap'])

    # Apply effective lower bound constraint
    if params.apply_elb:
        data['AdjustedTaylorRule'] = data['TaylorRule'].apply(
            lambda x: max(x, params.elb))
    else:
        data['AdjustedTaylorRule'] = data['TaylorRule']

    # Apply policy inertia
    data['AdjustedTaylorRule'] = (
            params.rho * data['FedRate'] +
            (1 - params.rho) * data['AdjustedTaylorRule'])

    return data.round(2)


def plot_historical_taylor_rule(
        historical_rule_estimates: pd.DataFrame
) -> None:
    """
    Extract the time range from the data, plot the Taylor Rule estimates and
    the Federal Funds Rate, and return the date range.

    Args:
        historical_rule_estimates (pd.DataFrame): DataFrame containing the
            historical rates including Taylor Rule estimates and the Federal
            Funds Rate.

    Returns:
        None
    """
    # Extracting the time range from the data
    start_date = historical_rule_estimates.dropna().index.min()
    end_date = historical_rule_estimates.dropna().index.max()
    date_range = (f"{start_date.strftime('%B %d, %Y')} to "
                  f"{end_date.strftime('%B %d, %Y')}")

    # Plotting Taylor Rule estimates and the Federal Funds Rate
    historical_rule_estimates[[
        'TaylorRule',
        'AdjustedTaylorRule',
        'FedRate'
    ]].dropna().plot(
        figsize=(10, 5),  # Specifies the figure size
        grid=True  # Enables grid lines for better readability
    )

    plt.title(f'Taylor Rule Estimates and Federal Funds Rate\n{date_range}')
    plt.xlabel('Year')
    plt.ylabel('Interest Rate (%)')
    plt.legend(['Taylor Rule', 'Adjusted Taylor Rule', 'Federal Funds Rate'])

    # Adding the citation as a footnote
    plt.figtext(
        x=0.25,
        y=-0.01,
        s="Data Source: Federal Reserve Economic Data (FRED)",
        ha="center"
    )

    plt.show()  # Display the plot



# pyeconomics/utils/bar_utils.py

from datetime import datetime


def verbose_balanced_approach_rule(data: dict):
    """
    Print verbose output for the Balanced Approach Rule calculation.

    Args:
        data (dict): Dictionary containing the following keys:
            - current_inflation_rate (float): Current inflation rate.
            - inflation_target (float): Target inflation rate.
            - current_unemployment_rate (float): Current unemployment rate.
            - natural_unemployment_rate (float): Natural unemployment rate.
            - long_term_real_interest_rate (float): Long-term real interest
              rate.
            - current_fed_rate (float): Current Federal Reserve rate.
            - inflation_gap (float): Inflation gap.
            - unemployment_gap (float): Unemployment gap.
            - unadjusted_rate (float): Unadjusted Balanced Approach Rule
              estimate.
            - adjusted_rate_after_elb (float): Balanced Approach Rule adjusted
              for the Effective Lower Bound (ELB).
            - adjusted_rate_after_inertia (float): Balanced Approach Rule
              adjusted for policy inertia.
            - rho (float): Policy inertia coefficient.
            - alpha (float): Coefficient for the inflation gap.
            - beta (float): Coefficient for the unemployment gap.
            - elb (float): Effective Lower Bound (ELB) rate.
            - apply_elb (bool): Whether to apply the ELB adjustment.
            - use_shortfalls_rule (bool): Whether to use the shortfalls rule.
            - use_shortfalls (bool): Whether to use the shortfalls gap.

    Returns:
        None
    """
    # Get the current date
    current_date = datetime.now().strftime("%B %d, %Y")

    # Determine if the shortfalls rule is being used
    rule_suffix = "Shortfalls " \
        if data.get('use_shortfalls_rule', False) else ""
    rule_acronym = f"BASR" if rule_suffix else "BAR"
    rule_name = f"Balanced Approach {rule_suffix}Rule (BASR)" \
        if rule_suffix else "Balanced Approach Rule"

    # Calculate the length of the line based on the rule name
    line_length = 67 + len(rule_acronym)

    print("\n==== Economic Indicators " + "=" * (line_length - 25))
    print("Current Inflation:                               {:.2f}%".format(
        data['current_inflation_rate']))
    print("Target Inflation:                                {:.2f}%".format(
        data['inflation_target']))
    print("Current Unemployment Rate:                       {:.2f}%".format(
        data['current_unemployment_rate']))
    print("Natural Unemployment Rate:                       {:.2f}%".format(
        data['natural_unemployment_rate']))
    print("Long-Term Real Interest Rate:                    {:.2f}%".format(
        data['long_term_real_interest_rate']))
    print("Current Fed Rate:                                {:.2f}%".format(
        data['current_fed_rate']))
    print("As of Date:                                      {}".format(
        current_date))
    print("\n==== Gaps " + "=" * (line_length - 10))
    print("Inflation Gap:                                   {:.2f}%".format(
        data['inflation_gap']))
    if data.get('use_shortfalls', False):
        print("Unemployment Shortfall Gap:                      {:.2f}%".format(
            data['unemployment_gap']))
    else:
        print("Unemployment Gap:                                {:.2f}%".format(
            data['unemployment_gap']))
    print("\n==== {} ".format(rule_name) + "=" *
          (line_length - 6 - len(rule_name)))
    print("  Long-Term Real Interest Rate:                  {:.2f}%".format(
        data['long_term_real_interest_rate']))
    print("  Current Inflation:                             + {:.2f}%".format(
        data['current_inflation_rate']))
    print("  Alpha * Inflation Gap:                         "
          "+ {:.2f} * {:.2f}%".format(
            data['alpha'], data['inflation_gap']))
    print("  Beta * Unemployment Gap:                       "
          "+ {:.2f} * {:.2f}%".format(
            data['beta'], data['unemployment_gap']))
    print("-" * line_length)
    print("  Unadjusted {} Estimate:".format(rule_acronym).ljust(49) +
          "{:.2f}%".format(data['unadjusted_rate']))
    print("\n==== Adjusted {} ".format(rule_name) + "=" *
          (line_length - 15 - len(rule_name)))
    if data['apply_elb']:
        print("  Effective Lower Bound (ELB) Adjustment: ")
        print("  Maximum of {} or ELB:".format(rule_acronym).ljust(49) +
              "max({:.2f}%, {:.2f}%)".format(
                data['unadjusted_rate'], data['elb']))
        print("-" * line_length)
        print("  {} Adjusted for ELB:".format(rule_acronym).ljust(49) +
              "{:.2f}%".format(data['adjusted_rate_after_elb']))
    print("\n  Policy Inertia Adjustment:              ")
    print("  Policy Inertia Coefficient (rho):              {:.2f}".format(
        data['rho']))
    print("  Current Fed Rate:                              * {:.2f}%".format(
        data['current_fed_rate']))
    print("  Adjustment Coefficient (1 - rho):              "
          "+ (1 - {:.2f})".format(data['rho']))
    print(f"  {rule_acronym} Adjusted for ELB:".ljust(49) +
          f"* {data['adjusted_rate_after_elb']:.2f}%")
    print("-" * line_length)
    print("  Adjusted {} Estimate:".format(rule_acronym).ljust(49) +
          "{:.2f}%".format(data['adjusted_rate_after_inertia']))

    # Policy Prescription section
    print("\n==== Policy Prescription " + "=" * (line_length - 25))
    rate_difference = (data['adjusted_rate_after_inertia'] -
                       data['current_fed_rate'])
    rounded_difference = round(rate_difference * 4) / 4

    if rounded_difference > 0.125:
        print(f"  The Adjusted {rule_acronym} Estimate is "
              f"{rate_difference:.2f}% higher than the Current Fed \n"
              f"  Rate. The Fed should consider raising the interest "
              f"rate by {rounded_difference:.2f}%.")
    elif rounded_difference < -0.125:
        print(f"  The Adjusted {rule_acronym} Estimate is "
              f"{abs(rate_difference):.2f}% lower than the Current Fed \n"
              f"  Rate. The Fed should consider lowering the interest "
              f"rate by {abs(rounded_difference):.2f}%.")
    else:
        print(f"  The Adjusted {rule_acronym} Estimate is equal to the Current "
              f"Fed Rate.\n  The Fed should maintain the current interest "
              f"rate.")

    # Adding a note explaining BAR and BASR
    print("\nNote:")
    print("-BAR stands for Balanced Approach Rule.")
    print("-BASR stands for Balanced Approach Shortfalls Rule.")



# pyeconomics/api/__init__.py

from .cache_manager import save_to_cache, load_from_cache
from .fred_api import FredClient, fred_client
from .fred_data import fetch_historical_fed_funds_rate

__all__ = ['FredClient', 'fred_client', 'fetch_historical_fed_funds_rate',
           'save_to_cache', 'load_from_cache']



# pyeconomics/utils/tr_utils.py

from datetime import datetime


def verbose_taylor_rule(data: dict):
    """
    Print verbose output for the Taylor Rule calculation.

    Args:
        data (dict): Dictionary containing the following keys:
            - current_inflation_rate (float): Current inflation rate.
            - inflation_target (float): Target inflation rate.
            - current_unemployment_rate (float): Current unemployment rate.
            - natural_unemployment_rate (float): Natural unemployment rate.
            - long_term_real_interest_rate (float): Long-term real interest
              rate.
            - current_fed_rate (float): Current Federal Reserve rate.
            - inflation_gap (float): Inflation gap.
            - unemployment_gap (float): Unemployment gap.
            - unadjusted_taylor_rule (float): Unadjusted Taylor Rule estimate.
            - adjusted_taylor_rule_after_elb (float): Taylor Rule adjusted for
                the Effective Lower Bound (ELB).
            - adjusted_taylor_rule_after_inertia (float): Taylor Rule adjusted
                for policy inertia.
            - rho (float): Policy inertia coefficient.
            - alpha (float): Coefficient for the inflation gap.
            - beta (float): Coefficient for the Okun factor multiplied by the
                unemployment gap.
            - okun_factor (float): Okun factor.
            - elb (float): Effective Lower Bound (ELB) rate.
            - apply_elb (bool): Whether to apply the ELB adjustment.

    Returns:
        None
    """
    current_date = datetime.now().strftime("%B %d, %Y")
    print("\n==== Economic Indicators ========================================="
          "========")
    print("Current Inflation:                               {:.2f}%".format(
        data['current_inflation_rate']))
    print("Target Inflation:                                {:.2f}%".format(
        data['inflation_target']))
    print("Current Unemployment Rate:                       {:.2f}%".format(
        data['current_unemployment_rate']))
    print("Natural Unemployment Rate:                       {:.2f}%".format(
        data['natural_unemployment_rate']))
    print("Long-Term Real Interest Rate:                    {:.2f}%".format(
        data['long_term_real_interest_rate']))
    print("Current Fed Rate:                                {:.2f}%".format(
        data['current_fed_rate']))
    print("As of Date:                                      {}".format(
        current_date))
    print("\n==== Gaps ========================================================"
          "========")
    print("Inflation Gap:                                   {:.2f}%".format(
        data['inflation_gap']))
    print("Unemployment Gap:                                {:.2f}%".format(
        data['unemployment_gap']))
    print("\n==== Taylor Rule ================================================="
          "========")
    print("  Long-Term Real Interest Rate:                  {:.2f}%".format(
        data['long_term_real_interest_rate']))
    print("  Current Inflation:                             + {:.2f}%".format(
        data['current_inflation_rate']))
    print("  Alpha * Inflation Gap:                         "
          "+ {:.2f} * {:.2f}%".format(
            data['alpha'], data['inflation_gap']))
    print("  Beta * Okun Factor * Unemployment Gap:         "
          "+ {:.2f} * {:.2f} * {:.2f}%".format(
            data['beta'], data['okun_factor'], data['unemployment_gap']))
    print("--------------------------------------------------------------------"
          "------")
    print("  Unadjusted Taylor Rule Estimate:               {:.2f}%".format(
        data['unadjusted_taylor_rule']))
    print("\n==== Adjusted Taylor Rule ========================================"
          "========")
    if data['apply_elb']:
        print("  Effective Lower Bound (ELB) Adjustment: ")
        print(
            "  Maximum of Taylor Rule or ELB:                 "
            "max({:.2f}%, {:.2f}%)".format(
                data['unadjusted_taylor_rule'], data['elb']))
        print("----------------------------------------------------------------"
              "----------")
        print("  Taylor Rule Adjusted for ELB:                  {:.2f}%".format(
            data['adjusted_taylor_rule_after_elb']))
    print("\n  Policy Inertia Adjustment:              ")
    print("  Policy Inertia Coefficient (rho):              {:.2f}".format(
        data['rho']))
    print("  Current Fed Rate:                              * {:.2f}%".format(
        data['current_fed_rate']))
    print("  Adjustment Coefficient (1 - rho):              "
          "+ (1 - {:.2f})".format(data['rho']))
    print("  Taylor Rule Adjusted for ELB:                  * {:.2f}%".format(
        data['adjusted_taylor_rule_after_elb']))
    print("--------------------------------------------------------------------"
          "------")
    print("  Adjusted Taylor Rule Estimate:                 {:.2f}%".format(
        data['adjusted_taylor_rule_after_inertia']))
    print("\n==== Policy Prescription ========================================="
          "========")
    rate_difference = (data['adjusted_taylor_rule_after_inertia'] -
                       data['current_fed_rate'])
    rounded_difference = round(rate_difference * 4) / 4

    if rounded_difference > 0.125:
        print("  The Adjusted Taylor Rule Estimate is {:.2f}% higher than the "
              "Current \n  Fed Rate. The Fed should consider raising the "
              "interest rate by {:.2f}%.".format(
                rate_difference, rounded_difference))
    elif rounded_difference < -0.125:
        print("  The Adjusted Taylor Rule Estimate is {:.2f}% lower than the "
              "Current \n  Fed Rate. The Fed should consider lowering the "
              "interest rate by {:.2f}%.".format(
                abs(rate_difference), rounded_difference))
    else:
        print("  The Adjusted Taylor Rule Estimate is equal to the Current "
              "Fed Rate.\n  The Fed should maintain the current interest rate.")



import os
import logging
import datetime
from threading import Lock
from typing import Optional, Any

import pandas as pd
from fredapi import Fred

from pyeconomics.api.cache_manager import save_to_cache, load_from_cache

try:
    import keyring
    KEYRING_AVAILABLE = True
except ImportError:
    keyring = None
    KEYRING_AVAILABLE = False


class DataSource:
    """
    Abstract base class for all data source clients.

    Attributes:
        api_key (Optional[str]): API key for accessing the data source.
    """

    def __init__(self, api_key: Optional[str] = None):
        self.api_key = api_key

    def fetch_data(self, series_id: str) -> pd.Series:
        """
        Abstract method to fetch data from a data source given a series ID.

        Args:
            series_id (str): The identifier for the data series.

        Returns:
            pandas.Series: Series containing the requested data.

        Raises:
            NotImplementedError: If method is not implemented.
        """
        raise NotImplementedError(
            "This method should be overridden by subclasses."
        )


class FredClient(DataSource):
    """
    A client for fetching data from the FRED API.

    Inherits from DataSource and implements methods to fetch data from FRED.
    """
    _instance: Optional['FredClient'] = None
    _lock: Lock = Lock()

    def __new__(cls, api_key: Optional[str] = None) -> 'FredClient':
        """
        Ensures a single instance of FredClient using a thread-safe singleton
        pattern.

        Args:
            api_key (Optional[str]): The FRED API key, retrieved from
                keyring if None.

        Returns:
            FredClient: Singleton instance.
        """
        with cls._lock:
            if cls._instance is None:
                logging.debug("Creating new FredClient instance")
                cls._instance = object.__new__(cls)
                api_key_retrieved = api_key or os.getenv('FRED_API_KEY')
                logging.debug(f"API key from environment variable: "
                              f"{api_key_retrieved}")
                if not api_key_retrieved and KEYRING_AVAILABLE:
                    logging.debug("Attempting to retrieve API key from keyring")
                    try:
                        api_key_retrieved = keyring.get_password(
                            "fred", "api_key")
                        logging.debug(f"API key from keyring: "
                                      f"{api_key_retrieved}")
                    except Exception as e:
                        logging.debug(f"Keyring not available: {e}")
                if not api_key_retrieved:
                    logging.debug("API Key not found, raising ValueError.")
                    raise ValueError(
                        "API Key for FRED must be provided "
                        "or retrievable from keyring.")
                logging.debug(f"Using API Key: {api_key_retrieved}")
                cls._instance.client = Fred(api_key=api_key_retrieved)
            return cls._instance

    @classmethod
    def reset_instance(cls) -> None:
        """
        Resets the singleton instance. Use this method only in tests.
        """
        with cls._lock:
            cls._instance = None
            logging.debug("FredClient instance reset")

    def fetch_data(self, series_id: str) -> pd.Series:
        """
        Fetches data for a given series ID from FRED with caching.

        Args:
            series_id (str): FRED series ID to fetch data for.

        Returns:
            pandas.Series: Series containing the requested data.

        Raises:
            ValueError: If no data is found for series ID.
            Exception: For fetch operation errors.
        """
        cache_key = f"fred_series_{series_id}"
        data = load_from_cache(cache_key)

        if data is not None:
            logging.info(f"Data for {series_id} loaded from cache.")
            return data

        try:
            data = self.client.get_series(series_id)
            if data.empty:
                raise ValueError(f"No data found for series ID {series_id}")
            save_to_cache(cache_key, data)
            logging.info(f"Data for {series_id} fetched and cached.")
            return data
        except Exception as e:
            logging.error(f"Fetching error for {series_id}: {e}")
            raise

    def get_latest_value(self, series_id: str) -> Optional[float]:
        """
        Fetches the latest value for a FRED series ID, considering only dates
        up to today.

        Args:
            series_id (str): Identifier for the FRED data series.

        Returns:
            Optional[float]: Most recent data point up to today, or None if no
                data.
        """
        try:
            data = self.fetch_data(series_id)
            today = datetime.date.today()
            filtered_data = data[:str(today)]
            return filtered_data.iloc[-1] if not filtered_data.empty else None
        except Exception as e:
            logging.error(f"Error in getting latest value for {series_id}: {e}")
            raise

    def get_historical_value(
        self,
        series_id: str,
        periods: int = -1
    ) -> Optional[float]:
        """
        Fetches a historical value for a FRED series ID.

        Args:
            series_id (str): Identifier for the FRED data series.
            periods (int): Index of period to retrieve (negative for
                historical).

        Returns:
            Optional[float]: Historical data point value, or None if
                unavailable.
        """
        try:
            series = self.fetch_data(series_id)
            return series.iloc[periods] \
                if not series.empty and len(series) > -periods else None
        except Exception as e:
            logging.error(
                f"Error in getting historical value for {series_id}: {e}")
            raise

    def get_series_name(self, series_id: str) -> str:
        """
        Fetches the name of a FRED series given its ID.

        Args:
            series_id (str): Identifier for the FRED data series.

        Returns:
            str: Name of the FRED series.
        """
        try:
            return self.client.get_series_info(series_id)["title"]
        except Exception as e:
            logging.error(f"Error in getting series name for {series_id}: {e}")
            raise

    @classmethod
    def get_data_or_fetch(
        cls, default: Any,
        series_id: str,
        periods: int = 0
    ) -> Any:
        """
        Fetches data if default is None using the given FRED series ID. Can
        fetch historical data based on periods.

        Args:
            default (Any): Default value to return if not None.
            series_id (str): FRED series ID for fetching data if default is
                None.
            periods (int): Number of periods back to fetch the data. Default is
                0 for current data.

        Returns:
            Any: Latest value or historical value or default value.
        """
        if default is not None:
            return default
        else:
            try:
                if periods == 0:
                    return cls._instance.get_latest_value(series_id)
                else:
                    return cls._instance.get_historical_value(
                        series_id, periods)
            except Exception as e:
                logging.error(
                    f"Error in get_data_or_fetch for {series_id}: {e}")
                raise


# Global instance of the FredClient
fred_client = FredClient()



# pyeconomics/data/economic_indicators.py

from dataclasses import dataclass

from typing import Optional


@dataclass
class EconomicIndicators:
    """
    Data class for storing economic indicators.

    Attributes:
        current_fed_rate (float, optional): Current Federal Funds Target Rate.
            Defaults to None.
        current_inflation_rate (float, optional): Current inflation rate.
            Defaults to None.
        current_unemployment_rate (float, optional): Current unemployment rate.
            Defaults to None.
        inflation_series_id (str): FRED Series ID for inflation data.
        lagged_natural_unemployment_rate (float, optional): The natural
            unemployment rate from a previous period. Defaults to None.
        lagged_unemployment_rate (float, optional): The unemployment rate from a
            previous period. Defaults to None.
        long_term_real_interest_rate (float, optional): Long-term real interest
            rate. Defaults to None.
        natural_unemployment_rate (float, optional): Natural unemployment rate.
            Defaults to None.
        natural_unemployment_series_id (str): FRED Series ID for natural
            unemployment rate.
        real_interest_rate_series_id (str): FRED Series ID for long-term real
            interest rate.
        unemployment_rate_series_id (str): FRED Series ID for unemployment rate.
    """
    current_fed_rate: Optional[float] = None
    current_inflation_rate: Optional[float] = None
    current_unemployment_rate: Optional[float] = None
    inflation_series_id: str = 'PCETRIM12M159SFRBDAL'
    lagged_natural_unemployment_rate: Optional[float] = None
    lagged_unemployment_rate: Optional[float] = None
    long_term_real_interest_rate: Optional[float] = None
    natural_unemployment_rate: Optional[float] = None
    natural_unemployment_series_id: str = 'NROU'
    real_interest_rate_series_id: str = 'DFII10'
    unemployment_rate_series_id: str = 'UNRATE'



# tests/test_fdr_utils.py

import pytest
from datetime import datetime

from pyeconomics.utils.fdr_utils import verbose_first_difference_rule


@pytest.fixture
def mock_data():
    return {
        'current_inflation_rate': 2.5,
        'inflation_target': 2.0,
        'current_unemployment_rate': 4.0,
        'lagged_unemployment_rate': 4.2,
        'natural_unemployment_rate': 4.5,
        'lagged_natural_unemployment_rate': 4.6,
        'current_fed_rate': 0.5,
        'inflation_gap': 0.5,
        'current_unemployment_gap': 0.5,
        'lagged_unemployment_gap': 0.4,
        'alpha': 0.5,
        'unadjusted_fdr_rule': 1.0,
        'adjusted_fdr_rule_after_elb': 1.0,
        'apply_elb': False,
        'elb': 0.0,
        'rho': 0.5,
        'adjusted_fdr_rule_after_inertia': 0.75
    }


def test_verbose_first_difference_rule(capsys, mock_data):
    verbose_first_difference_rule(mock_data)

    captured = capsys.readouterr()
    output = captured.out

    as_of_date = datetime.now().strftime("%B %d, %Y")
    expected_output = (
        "\n==== Economic Indicators "
        "============================================\n"
        "Current Inflation:                               2.50%\n"
        "Target Inflation:                                2.00%\n"
        "Current Unemployment Rate:                       4.00%\n"
        "Lagged Unemployment Rate:                        4.20%\n"
        "Natural Unemployment Rate:                       4.50%\n"
        "Lagged Natural Unemployment Rate:                4.60%\n"
        "Last Fed Rate:                                   0.50%\n"
        f"As of Date:                                      {as_of_date}\n"
        "\n==== Gaps "
        "===========================================================\n"
        "Inflation Gap:                                   0.50%\n"
        "Current Unemployment Gap:                        0.50%\n"
        "Lagged Unemployment Gap:                         0.40%\n"
        "\n==== First Difference Rule (FDR) "
        "====================================\n"
        "  Last Fed Rate:                                 0.50%\n"
        "  Alpha * Inflation Gap:                         + 0.50 * 0.50%\n"
        "  Current Unemployment Gap Adjustment:           + 0.50%\n"
        "  Last Year Unemployment Gap Adjustment:         - 0.40%\n"
        "-----------------------------------"
        "----------------------------------\n"
        "  Unadjusted FDR Estimate:                       1.00%\n"
        "\n==== Adjusted First Difference Rule "
        "=================================\n"
        "\n  Policy Inertia Adjustment:\n"
        "  Policy Inertia Coefficient (rho):              0.50\n"
        "  Last Fed Rate:                                 * 0.50%\n"
        "  Adjustment Coefficient (1 - rho):              + (1 - 0.50)\n"
        "  FDR Adjusted for ELB:                          * 1.00%\n"
        "-----------------------------------"
        "----------------------------------\n"
        "  Adjusted FDR Estimate:                         0.75%\n"
        "\n==== Policy Prescription "
        "=================================================\n"
        "  The Adjusted FDR Estimate is 0.25% higher than the Current Fed \n"
        "  Rate. The Fed should consider raising the interest rate by 0.25%.\n"
    )

    # Comparing the captured output with the expected output
    assert output == expected_output


def test_verbose_first_difference_rule_lower_rate(capsys, mock_data):
    mock_data['adjusted_fdr_rule_after_inertia'] = 0.0
    verbose_first_difference_rule(mock_data)

    captured = capsys.readouterr()
    output = captured.out

    assert ("The Adjusted FDR Estimate is 0.50% "
            "lower than the Current Fed") in output
    assert ("Rate. The Fed should consider lowering "
            "the interest rate by 0.50%.") in output


def test_verbose_first_difference_rule_equal_rate(capsys, mock_data):
    mock_data['adjusted_fdr_rule_after_inertia'] = 0.5
    verbose_first_difference_rule(mock_data)

    captured = capsys.readouterr()
    output = captured.out

    assert ("The Adjusted FDR Estimate is equal"
            " to the Current Fed Rate.") in output
    assert "The Fed should maintain the current interest rate." in output


if __name__ == '__main__':
    pytest.main()



# pyeconomics/data/model_parameters.py

from dataclasses import dataclass


@dataclass
class TaylorRuleParameters:
    """
    Data class for storing Taylor Rule model parameters.

    Attributes:
        inflation_target (float): Target inflation rate.
        alpha (float): Weight for inflation gap.
        beta (float): Weight for unemployment gap.
        okun_factor (float): Multiplier for the unemployment gap.
        rho (float): Policy inertia coefficient.
        elb (float): Effective lower bound for interest rates.
        apply_elb (bool): Whether to apply the effective lower bound
            constraint to the Taylor Rule estimate.
        verbose (bool): Whether to print verbose output.
    """
    inflation_target: float = 2.0
    alpha: float = 0.5
    beta: float = 0.5
    okun_factor: float = 2.0
    rho: float = 0.0
    elb: float = 0.125
    apply_elb: bool = False
    verbose: bool = False


@dataclass
class BalancedApproachRuleParameters:
    """
    Data class for storing Balanced Approach Rule parameters.

    Attributes:
        inflation_target (float): Target inflation rate.
        alpha (float): Weight for inflation gap.
        beta (float): Weight for unemployment gap.
        rho (float): Policy inertia coefficient.
        elb (float): Effective lower bound.
        apply_elb (bool): Whether to apply the effective lower bound.
        use_shortfalls_rule (bool): Whether to use the Balanced Approach Rule
            with shortfalls.
        verbose (bool): Whether to print verbose output.
    """
    inflation_target: float = 2.0
    alpha: float = 0.5
    beta: float = 2.0
    rho: float = 0.0
    elb: float = 0.125
    apply_elb: bool = False
    use_shortfalls_rule: bool = False
    verbose: bool = False


@dataclass
class FirstDifferenceRuleParameters:
    """
    Data class for storing First Difference Rule parameters.

    Attributes:
        inflation_target (float): Long-term target inflation rate.
        alpha (float): Coefficient for the inflation gap.
        rho (float): Policy inertia coefficient.
        elb (float): Effective lower bound for the interest rate.
        apply_elb (bool): Whether to apply the effective lower bound constraint.
        verbose (bool): Whether to print verbose output.
    """
    inflation_target: float = 2.0
    alpha: float = 0.5
    rho: float = 0.0
    elb: float = 0.125
    apply_elb: bool = False
    verbose: bool = False



# tests/test_fred_data.py

from unittest.mock import patch

import pandas as pd
import pytest

from pyeconomics.api.fred_data import fetch_historical_fed_funds_rate


@pytest.fixture
def mock_fred_data():
    """Fixture for mock FRED data."""
    dfedtar_data = {
        'DATE': pd.date_range(start='2000-01-01', end='2008-12-15', freq='ME'),
        'VALUE': [1.5] * 108
    }
    dfedtaru_data = {
        'DATE': pd.date_range(start='2008-12-16', end='2020-01-01', freq='ME'),
        'VALUE': [0.25] * 133
    }
    # Fix the lengths to ensure they match
    dfedtar_data['VALUE'] = dfedtar_data['VALUE'][:len(dfedtar_data['DATE'])]
    dfedtaru_data['VALUE'] = dfedtaru_data['VALUE'][:len(dfedtaru_data['DATE'])]

    dfedtar = pd.DataFrame(dfedtar_data).set_index('DATE')
    dfedtaru = pd.DataFrame(dfedtaru_data).set_index('DATE')
    return dfedtar, dfedtaru


@patch('pyeconomics.api.fred_data.fred_client')
def test_fetch_historical_fed_funds_rate(mock_fred_client, mock_fred_data):
    dfedtar, dfedtaru = mock_fred_data
    mock_fred_client.fetch_data.side_effect = [dfedtar, dfedtaru]

    result = fetch_historical_fed_funds_rate()

    expected = pd.concat([
        dfedtar[dfedtar.index <= '2008-12-15'],
        dfedtaru[dfedtaru.index > '2008-12-15']
    ])
    expected.index.name = 'FedRate'
    expected.name = 'FedRate'

    pd.testing.assert_frame_equal(result, expected)


if __name__ == '__main__':
    pytest.main()



# tests/test_balanced_approach_rule.py

from unittest.mock import patch
import pandas as pd
import pytest

from pyeconomics.data.economic_indicators import EconomicIndicators
from pyeconomics.data.model_parameters import BalancedApproachRuleParameters

from pyeconomics.models.monetary_policy.balanced_approach_rule import (
    balanced_approach_rule,
    historical_balanced_approach_rule,
    plot_historical_bar_basr_rule
)


@pytest.fixture
def mock_fred_data():
    """Fixture for mock FRED data."""
    return {
        'inflation_rate': 2.5,
        'unemployment_rate': 4.0,
        'natural_unemployment_rate': 4.5,
        'real_interest_rate': 1.0,
        'current_fed_rate': 0.5
    }


@patch('pyeconomics.models.monetary_policy.balanced_approach_rule.fred_client')
def test_balanced_approach_rule(mock_fred_client, mock_fred_data):
    mock_fred_client.get_data_or_fetch.side_effect = (
        lambda default, series_id, periods=0: mock_fred_data.get(series_id)
    )
    mock_fred_client.get_latest_value.return_value = (
        mock_fred_data['current_fed_rate']
    )

    indicators = EconomicIndicators(
        inflation_series_id='inflation_rate',
        unemployment_rate_series_id='unemployment_rate',
        natural_unemployment_series_id='natural_unemployment_rate',
        real_interest_rate_series_id='real_interest_rate'
    )

    params = BalancedApproachRuleParameters(
        inflation_target=2.0, alpha=0.5, beta=2.0,
        rho=0.0, elb=0.125, apply_elb=False,
        use_shortfalls_rule=False, verbose=False
    )

    # Default test case
    result = balanced_approach_rule(indicators, params)
    expected_result = 4.75  # Computed based on the mock data and parameters
    assert result == expected_result

    # Test with apply_elb = True
    params.apply_elb = True
    result = balanced_approach_rule(indicators, params)
    expected_result = max(4.75, 0.125)
    assert result == expected_result

    # Test with use_shortfalls_rule = True
    params.use_shortfalls_rule = True
    params.apply_elb = False
    result = balanced_approach_rule(indicators, params)
    expected_result = 3.75  # Change as the gap is negative (shortfall)
    assert result == expected_result

    # Test with verbose output
    with patch('pyeconomics.models.monetary_policy.balanced_approach_rule.'
               'verbose_balanced_approach_rule') as mock_verbose:
        params.verbose = True
        balanced_approach_rule(indicators, params)
        assert mock_verbose.called

    # Test ValueError raise condition
    with pytest.raises(ValueError):
        invalid_indicators = EconomicIndicators(
            inflation_series_id='missing_series_id',
            unemployment_rate_series_id='missing_series_id',
            natural_unemployment_series_id='missing_series_id',
            real_interest_rate_series_id='missing_series_id'
        )
        balanced_approach_rule(invalid_indicators, params)

    # Test with current_fed_rate as None
    mock_fred_client.get_latest_value.return_value = 1.0
    indicators.current_fed_rate = None
    params.verbose = False
    result = balanced_approach_rule(indicators, params)
    expected_result = 3.75  # Computed based on the mock data and parameters
    assert result == expected_result

    # Test with current_fed_rate provided
    indicators.current_fed_rate = 2.0
    result = balanced_approach_rule(indicators, params)
    expected_result = 3.75  # Computed based on the mock data and parameters
    assert result == expected_result


@patch('pyeconomics.models.monetary_policy.balanced_approach_rule.fred_client')
@patch(
    'pyeconomics.models.monetary_policy.'
    'balanced_approach_rule.fetch_historical_fed_funds_rate'
)
def test_historical_balanced_approach_rule(
    mock_fetch_historical_fed_funds_rate,
    mock_fred_client,
    mock_fred_data
):
    mock_fred_client.fetch_data.side_effect = (
        lambda series_id: pd.Series([mock_fred_data[series_id]] * 10)
    )
    mock_fetch_historical_fed_funds_rate.return_value = (
        pd.Series([mock_fred_data['current_fed_rate']] * 10)
    )

    indicators = EconomicIndicators(
        inflation_series_id='inflation_rate',
        unemployment_rate_series_id='unemployment_rate',
        natural_unemployment_series_id='natural_unemployment_rate',
        real_interest_rate_series_id='real_interest_rate'
    )

    params = BalancedApproachRuleParameters(
        inflation_target=2.0, alpha=0.5, beta=2.0,
        rho=0.0, elb=0.125, apply_elb=False,
        use_shortfalls_rule=False, verbose=False
    )

    # Default test case
    result = historical_balanced_approach_rule(indicators, params)

    # Verify the shape and content of the result DataFrame
    assert not result.empty
    assert all(result['BalancedApproachRule'] == 4.75)  # Based on mock data

    # Test with apply_elb = True
    params.apply_elb = True
    result = historical_balanced_approach_rule(indicators, params)
    assert not result.empty
    assert all(result['AdjustedBalancedApproachRule'] == max(4.75, 0.125))

    # Test with use_shortfalls_rule = True
    params.apply_elb = False
    params.use_shortfalls_rule = True
    result = historical_balanced_approach_rule(indicators, params)
    assert not result.empty
    assert all(
        result['BalancedApproachShortfallsRule'] == 3.75)  # Based on mock data


@patch('pyeconomics.models.monetary_policy.balanced_approach_rule.plt.show')
def test_plot_historical_bar_basr_rule(mock_show):
    historical_rates = pd.DataFrame({
        'BalancedApproachRule': [3.0, 3.1],
        'BalancedApproachShortfallsRule': [2.7, 2.8],
        'AdjustedBalancedApproachRule': [3.2, 3.3],
        'AdjustedBalancedApproachShortfallsRule': [2.9, 3.0],
        'FedRate': [2.0, 2.1]
    }, index=pd.to_datetime(['2020-01-01', '2020-02-01']))

    # Test unadjusted plot
    plot_historical_bar_basr_rule(historical_rates, adjusted=False)
    assert mock_show.called

    # Test adjusted plot
    plot_historical_bar_basr_rule(historical_rates, adjusted=True)
    assert mock_show.called


if __name__ == '__main__':
    pytest.main()



# tests/test_monetary_policy_rules.py

import unittest
from unittest.mock import patch

import pandas as pd

from pyeconomics.data.economic_indicators import EconomicIndicators

from pyeconomics.models.monetary_policy.monetary_policy_rules import (
    print_fred_series_names,
    print_verbose_output,
    calculate_policy_rule_estimates,
    calculate_historical_policy_rates,
    plot_historical_rule_estimates
)


class TestMonetaryPolicyRules(unittest.TestCase):

    @patch(
        'pyeconomics.models.monetary_policy.'
        'monetary_policy_rules.fred_client.get_series_name')
    def test_print_fred_series_names(self, mock_get_series_name):
        mock_get_series_name.side_effect = lambda x: f"Mocked Name for {x}"
        print_fred_series_names()
        self.assertEqual(mock_get_series_name.call_count, 4)

    @patch('pyeconomics.models.monetary_policy.monetary_policy_rules.datetime')
    def test_print_verbose_output(self, mock_datetime):
        mock_datetime.now.return_value.strftime.return_value = "May 20, 2024"
        estimates = pd.DataFrame({
            'Estimate (%)': [2.5, 3.0, 3.5, 2.0]
        }, index=[
            'Taylor Rule (TR)',
            'Balanced Approach Rule (BAR)',
            'Balanced Approach Shortfalls Rule (BASR)',
            'First Difference Rule (FDR)'
        ])
        print_verbose_output(estimates, 2.0)
        self.assertTrue(True)  # No exception should be raised

        print_verbose_output(estimates, 2.0, adjusted=True)
        self.assertTrue(True)  # No exception should be raised

    @patch(
        'pyeconomics.models.monetary_policy.monetary_policy_rules.fred_client.'
        'get_latest_value')
    @patch(
        'pyeconomics.models.monetary_policy.monetary_policy_rules.taylor_rule')
    @patch(
        'pyeconomics.models.monetary_policy.monetary_policy_rules.'
        'balanced_approach_rule')
    @patch(
        'pyeconomics.models.monetary_policy.monetary_policy_rules.'
        'first_difference_rule')
    def test_calculate_policy_rule_estimates(
        self, mock_first_difference_rule, mock_balanced_approach_rule,
            mock_taylor_rule, mock_get_latest_value):
        mock_get_latest_value.return_value = 2.0
        mock_taylor_rule.return_value = 2.5
        mock_balanced_approach_rule.return_value = 3.0
        mock_first_difference_rule.return_value = 1.5

        indicators = EconomicIndicators(
            inflation_series_id='inflation_rate',
            unemployment_rate_series_id='unemployment_rate',
            natural_unemployment_series_id='natural_unemployment_rate',
            real_interest_rate_series_id='real_interest_rate'
        )

        estimates = calculate_policy_rule_estimates(indicators, verbose=True)
        self.assertIsInstance(estimates, pd.DataFrame)
        self.assertEqual(estimates.shape, (4, 1))

    @patch(
        'pyeconomics.models.monetary_policy.monetary_policy_rules.'
        'fetch_historical_fed_funds_rate')
    @patch(
        'pyeconomics.models.monetary_policy.monetary_policy_rules.'
        'historical_taylor_rule')
    @patch(
        'pyeconomics.models.monetary_policy.monetary_policy_rules.'
        'historical_balanced_approach_rule')
    @patch(
        'pyeconomics.models.monetary_policy.monetary_policy_rules.'
        'historical_first_difference_rule')
    def test_calculate_historical_policy_rates(
        self, mock_historical_first_difference_rule,
        mock_historical_balanced_approach_rule,
            mock_historical_taylor_rule, mock_fetch_historical_fed_funds_rate):
        mock_historical_taylor_rule.return_value = pd.DataFrame({
            'TaylorRule': [2.5, 2.6],
            'AdjustedTaylorRule': [2.7, 2.8]
        }, index=pd.to_datetime(['2020-01-01', '2020-02-01']))

        mock_historical_balanced_approach_rule.return_value = pd.DataFrame({
            'BalancedApproachRule': [3.0, 3.1],
            'BalancedApproachShortfallsRule': [2.7, 2.8],
            'AdjustedBalancedApproachRule': [3.2, 3.3],
            'AdjustedBalancedApproachShortfallsRule': [2.9, 3.0]
        }, index=pd.to_datetime(['2020-01-01', '2020-02-01']))

        mock_historical_first_difference_rule.return_value = pd.DataFrame({
            'FirstDifferenceRule': [1.5, 1.6],
            'AdjustedFirstDifferenceRule': [1.7, 1.8]
        }, index=pd.to_datetime(['2020-01-01', '2020-02-01']))

        mock_fetch_historical_fed_funds_rate.return_value = pd.DataFrame({
            'FedRate': [2.0, 2.1]
        }, index=pd.to_datetime(['2020-01-01', '2020-02-01']))

        indicators = EconomicIndicators(
            inflation_series_id='inflation_rate',
            unemployment_rate_series_id='unemployment_rate',
            natural_unemployment_series_id='natural_unemployment_rate',
            real_interest_rate_series_id='real_interest_rate'
        )

        historical_rates = calculate_historical_policy_rates(
            indicators, rho=0.7, apply_elb=True
        )
        self.assertIsInstance(historical_rates, pd.DataFrame)
        self.assertEqual(historical_rates.shape,
                         (2, 9))  # Adjusted for new columns

    @patch('pyeconomics.models.monetary_policy.monetary_policy_rules.plt.show')
    def test_plot_historical_policy_rates(self, mock_show):
        historical_policy_rates = pd.DataFrame({
            'TaylorRule': [2.5, 2.6],
            'BalancedApproachRule': [3.0, 3.1],
            'BalancedApproachShortfallsRule': [2.7, 2.8],
            'FirstDifferenceRule': [1.5, 1.6],
            'FedRate': [2.0, 2.1],
            'AdjustedTaylorRule': [2.7, 2.8],
            'AdjustedBalancedApproachRule': [3.2, 3.3],
            'AdjustedBalancedApproachShortfallsRule': [2.9, 3.0],
            'AdjustedFirstDifferenceRule': [1.7, 1.8]
        }, index=pd.to_datetime(['2020-01-01', '2020-02-01']))

        # Test unadjusted plot
        plot_historical_rule_estimates(historical_policy_rates, adjusted=False)
        self.assertTrue(mock_show.called)

        # Test adjusted plot
        plot_historical_rule_estimates(historical_policy_rates, adjusted=True)
        self.assertTrue(mock_show.called)

    @patch(
        'pyeconomics.models.monetary_policy.monetary_policy_rules.'
        'print_verbose_output')
    @patch(
        'pyeconomics.models.monetary_policy.monetary_policy_rules.fred_client.'
        'get_latest_value')
    @patch(
        'pyeconomics.models.monetary_policy.monetary_policy_rules.taylor_rule')
    @patch(
        'pyeconomics.models.monetary_policy.monetary_policy_rules.'
        'balanced_approach_rule')
    @patch(
        'pyeconomics.models.monetary_policy.monetary_policy_rules.'
        'first_difference_rule')
    def test_calculate_policy_rule_estimates_with_none_fed_rate(
        self, mock_first_difference_rule, mock_balanced_approach_rule,
            mock_taylor_rule, mock_get_latest_value, mock_print_verbose_output):
        mock_get_latest_value.return_value = 2.0
        mock_taylor_rule.return_value = 2.5
        mock_balanced_approach_rule.return_value = 3.0
        mock_first_difference_rule.return_value = 1.5

        indicators = EconomicIndicators(
            inflation_series_id='inflation_rate',
            unemployment_rate_series_id='unemployment_rate',
            natural_unemployment_series_id='natural_unemployment_rate',
            real_interest_rate_series_id='real_interest_rate'
        )

        # Test with current_fed_rate as None
        estimates = calculate_policy_rule_estimates(
            indicators, verbose=True, rho=0.7, apply_elb=True
        )
        self.assertIsInstance(estimates, pd.DataFrame)
        self.assertEqual(estimates.shape, (4, 1))
        self.assertTrue(mock_print_verbose_output.called)

        # Test verbose only
        estimates = calculate_policy_rule_estimates(
            indicators, verbose=True, rho=0.0, apply_elb=False
        )
        self.assertIsInstance(estimates, pd.DataFrame)
        self.assertEqual(estimates.shape, (4, 1))
        self.assertTrue(mock_print_verbose_output.called)

    @patch(
        'pyeconomics.models.monetary_policy.monetary_policy_rules.'
        'print_verbose_output')
    @patch(
        'pyeconomics.models.monetary_policy.monetary_policy_rules.fred_client.'
        'get_latest_value')
    @patch(
        'pyeconomics.models.monetary_policy.monetary_policy_rules.taylor_rule')
    @patch(
        'pyeconomics.models.monetary_policy.monetary_policy_rules.'
        'balanced_approach_rule')
    @patch(
        'pyeconomics.models.monetary_policy.monetary_policy_rules.'
        'first_difference_rule')
    def test_calculate_policy_rule_estimates_verbose_only(
        self, mock_first_difference_rule, mock_balanced_approach_rule,
            mock_taylor_rule, mock_get_latest_value, mock_print_verbose_output):
        mock_get_latest_value.return_value = 2.0
        mock_taylor_rule.return_value = 2.5
        mock_balanced_approach_rule.return_value = 3.0
        mock_first_difference_rule.return_value = 1.5

        indicators = EconomicIndicators(
            inflation_series_id='inflation_rate',
            unemployment_rate_series_id='unemployment_rate',
            natural_unemployment_series_id='natural_unemployment_rate',
            real_interest_rate_series_id='real_interest_rate'
        )

        # Test with verbose only (rho=0.0 and apply_elb=False)
        estimates = calculate_policy_rule_estimates(
            indicators, verbose=True, rho=0.0, apply_elb=False
        )
        self.assertIsInstance(estimates, pd.DataFrame)
        self.assertEqual(estimates.shape, (4, 1))
        self.assertTrue(mock_print_verbose_output.called)

        # Reset mock call count before next test
        mock_print_verbose_output.reset_mock()

        # Test with verbose set to False
        estimates = calculate_policy_rule_estimates(
            indicators, verbose=False, rho=0.0, apply_elb=False
        )
        self.assertIsInstance(estimates, pd.DataFrame)
        self.assertEqual(estimates.shape, (4, 1))
        self.assertFalse(mock_print_verbose_output.called)


if __name__ == '__main__':
    unittest.main()



import os
import re
from setuptools import setup, find_packages


# Function to extract the version from __version__.py
def get_version():
    here = os.path.abspath(os.path.dirname(__file__))
    with (open(
        os.path.join(here, '__version__.py'),
        encoding='utf-8'
    ) as version_file):
        version_file_contents = version_file.read()
    # Use regular expression to extract version string
    version_match = re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]",
        version_file_contents,
        re.M
    )
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


# Read the contents of your requirements file
with open('requirements.txt', encoding='utf-8') as req_file:
    required = req_file.read().splitlines()

# Open and read the README file for the long description
with open('README.md', encoding='utf-8') as readme_file:
    long_description = readme_file.read()

setup(
    name='pyeconomics',
    version=get_version(),
    author='Nathan Ramos, CFA',
    author_email='nathan.ramos.github@gmail.com',
    packages=find_packages(),
    include_package_data=True,
    description='A library for economic and financial analysis',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/nathanramoscfa/pyeconomics',
    install_requires=required,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Financial and Insurance Industry',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Office/Business :: Financial :: Investment'
    ],
    python_requires='>=3.11, <3.13',
    project_urls={
        'Bug Reports': 'https://github.com/nathanramoscfa/pyeconomics/issues',
        'Source': 'https://github.com/nathanramoscfa/pyeconomics',
    },
)



# tests/test_first_difference_rule.py

import pytest
import pandas as pd
from unittest.mock import patch

from pyeconomics.models.monetary_policy.first_difference_rule import (
    first_difference_rule,
    historical_first_difference_rule,
    plot_historical_fdr
)
from pyeconomics.api import FredClient
from pyeconomics.data.economic_indicators import EconomicIndicators
from pyeconomics.data.model_parameters import FirstDifferenceRuleParameters


@pytest.fixture
def mock_fred_client():
    with patch.object(FredClient, 'fetch_data') as mock_fetch_data:
        yield mock_fetch_data


@pytest.fixture
def sample_fred_data():
    data = {
        '2023-01-01': 2.0,
        '2023-02-01': 2.1,
        '2023-03-01': 2.2,
        '2023-04-01': 2.3,
        '2023-05-01': 2.4,
        '2023-06-01': 2.5,
        '2023-07-01': 2.6,
        '2023-08-01': 2.7,
        '2023-09-01': 2.8,
        '2023-10-01': 2.9,
        '2023-11-01': 3.0,
        '2023-12-01': 3.1,
    }
    return pd.Series(data)


def test_first_difference_rule(mock_fred_client, sample_fred_data):
    mock_fred_client.return_value = sample_fred_data

    indicators = EconomicIndicators(
        current_inflation_rate=3.0,
        current_unemployment_rate=4.0,
        natural_unemployment_rate=3.5,
        lagged_unemployment_rate=4.2,
        lagged_natural_unemployment_rate=3.6,
        current_fed_rate=2.5
    )

    params = FirstDifferenceRuleParameters(
        inflation_target=2.0,
        alpha=0.5,
        rho=0.5,
        elb=0.125,
        apply_elb=True,
        verbose=False
    )

    rate = first_difference_rule(indicators, params)

    assert isinstance(rate, float)
    assert rate == 2.8  # Adjusted expected value based on actual calculation

    params.apply_elb = False
    rate = first_difference_rule(indicators, params)

    assert isinstance(rate, float)
    assert rate == 2.8  # Value without applying ELB

    mock_fred_client.get_latest_value.return_value = 1.0
    indicators.current_fed_rate = None
    rate = first_difference_rule(indicators, params)

    assert isinstance(rate, float)
    assert rate == 3.4  # Adjusted expected value based on actual calculation


def test_historical_first_difference_rule(mock_fred_client, sample_fred_data):
    mock_fred_client.side_effect = lambda series_id: sample_fred_data

    indicators = EconomicIndicators(
        inflation_series_id='PCETRIM12M159SFRBDAL',
        unemployment_rate_series_id='UNRATE',
        natural_unemployment_series_id='NROU'
    )

    params = FirstDifferenceRuleParameters(
        inflation_target=2.0,
        alpha=0.5,
        rho=0.5,
        elb=0.125,
        apply_elb=True
    )

    historical_rates = historical_first_difference_rule(indicators, params)

    assert isinstance(historical_rates, pd.DataFrame)
    assert 'AdjustedFirstDifferenceRule' in historical_rates.columns
    assert historical_rates['AdjustedFirstDifferenceRule'].notnull().all()

    params.apply_elb = False
    historical_rates = historical_first_difference_rule(indicators, params)

    assert isinstance(historical_rates, pd.DataFrame)
    assert 'AdjustedFirstDifferenceRule' in historical_rates.columns
    assert historical_rates['AdjustedFirstDifferenceRule'].notnull().all()


def test_historical_first_difference_rule_missing_data(
    mock_fred_client, sample_fred_data
):
    mock_fred_client.side_effect = [
        sample_fred_data, sample_fred_data, sample_fred_data,
        sample_fred_data, pd.Series()
    ]

    indicators = EconomicIndicators(
        inflation_series_id='PCETRIM12M159SFRBDAL',
        unemployment_rate_series_id='UNRATE',
        natural_unemployment_series_id='NROU'
    )

    params = FirstDifferenceRuleParameters(
        inflation_target=2.0,
        alpha=0.5,
        rho=0.5,
        elb=0.125,
        apply_elb=True
    )

    with pytest.raises(ValueError):
        historical_first_difference_rule(indicators, params)


def test_first_difference_rule_verbose_output(
    mock_fred_client,
    sample_fred_data,
    capsys
):
    mock_fred_client.return_value = sample_fred_data

    indicators = EconomicIndicators(
        current_inflation_rate=3.0,
        current_unemployment_rate=4.0,
        natural_unemployment_rate=3.5,
        lagged_unemployment_rate=4.2,
        lagged_natural_unemployment_rate=3.6,
        current_fed_rate=2.5
    )

    params = FirstDifferenceRuleParameters(
        inflation_target=2.0,
        alpha=0.5,
        rho=0.5,
        elb=0.125,
        apply_elb=True,
        verbose=True
    )

    first_difference_rule(indicators, params)

    captured = capsys.readouterr()
    assert "==== First Difference Rule (FDR) ===" in captured.out


def test_first_difference_rule_value_error(mock_fred_client, sample_fred_data):
    mock_fred_client.side_effect = lambda default, series_id, periods=0: None

    indicators = EconomicIndicators(
        inflation_series_id='missing_series_id',
        unemployment_rate_series_id='missing_series_id',
        natural_unemployment_series_id='missing_series_id'
    )

    params = FirstDifferenceRuleParameters(
        inflation_target=2.0,
        alpha=0.5,
        rho=0.0,
        elb=0.125,
        apply_elb=False,
        verbose=False
    )

    with pytest.raises(ValueError):
        first_difference_rule(indicators, params)


def test_historical_first_difference_rule_exception(
    mock_fred_client, sample_fred_data
):
    mock_fred_client.side_effect = lambda series_id: None

    indicators = EconomicIndicators(
        inflation_series_id='PCETRIM12M159SFRBDAL',
        unemployment_rate_series_id='UNRATE',
        natural_unemployment_series_id='NROU'
    )

    params = FirstDifferenceRuleParameters(
        inflation_target=2.0,
        alpha=0.5,
        rho=0.5,
        elb=0.125,
        apply_elb=True
    )

    with pytest.raises(ValueError):
        historical_first_difference_rule(indicators, params)


def test_historical_first_difference_rule_apply_elb(
    mock_fred_client, sample_fred_data
):
    mock_fred_client.side_effect = lambda series_id: sample_fred_data

    indicators = EconomicIndicators(
        inflation_series_id='PCETRIM12M159SFRBDAL',
        unemployment_rate_series_id='UNRATE',
        natural_unemployment_series_id='NROU'
    )

    params = FirstDifferenceRuleParameters(
        inflation_target=2.0,
        alpha=0.5,
        rho=0.0,
        elb=0.125,
        apply_elb=True
    )

    historical_rates = historical_first_difference_rule(indicators, params)

    assert isinstance(historical_rates, pd.DataFrame)
    assert 'AdjustedFirstDifferenceRule' in historical_rates.columns
    assert historical_rates['AdjustedFirstDifferenceRule'].notnull().all()
    assert (historical_rates['AdjustedFirstDifferenceRule'] >= 0.125).all()


def test_historical_first_difference_rule_partial_data(
    mock_fred_client, sample_fred_data
):
    # Mock to return partial data (None for one series)
    mock_fred_client.side_effect = [
        sample_fred_data, sample_fred_data, sample_fred_data, None,
        sample_fred_data
    ]

    indicators = EconomicIndicators(
        inflation_series_id='PCETRIM12M159SFRBDAL',
        unemployment_rate_series_id='UNRATE',
        natural_unemployment_series_id='NROU'
    )

    params = FirstDifferenceRuleParameters(
        inflation_target=2.0,
        alpha=0.5,
        rho=0.5,
        elb=0.125,
        apply_elb=True
    )

    with pytest.raises(ValueError):
        historical_first_difference_rule(indicators, params)


def test_historical_first_difference_rule_apply_elb_lambda(
    mock_fred_client, sample_fred_data
):
    # Mock to return full data
    mock_fred_client.side_effect = lambda series_id: sample_fred_data

    indicators = EconomicIndicators(
        inflation_series_id='PCETRIM12M159SFRBDAL',
        unemployment_rate_series_id='UNRATE',
        natural_unemployment_series_id='NROU'
    )

    params = FirstDifferenceRuleParameters(
        inflation_target=2.0,
        alpha=0.5,
        rho=0.5,
        elb=0.125,
        apply_elb=True
    )

    historical_rates = historical_first_difference_rule(indicators, params)

    assert isinstance(historical_rates, pd.DataFrame)
    assert 'AdjustedFirstDifferenceRule' in historical_rates.columns
    assert historical_rates['AdjustedFirstDifferenceRule'].notnull().all()
    assert all(
        a == b
        for a, b in zip(
            historical_rates['AdjustedFirstDifferenceRule'],
            historical_rates['FirstDifferenceRule'].apply(
                lambda x: max(x, 0.125))
        )
    )


@patch('pyeconomics.models.monetary_policy.first_difference_rule.plt.show')
def test_plot_historical_fdr(mock_show):
    historical_rates = pd.DataFrame({
        'FirstDifferenceRule': [2.5, 2.6],
        'AdjustedFirstDifferenceRule': [2.7, 2.8],
        'FedRate': [2.0, 2.1]
    }, index=pd.to_datetime(['2020-01-01', '2020-02-01']))

    # Test plot
    plot_historical_fdr(historical_rates)
    assert mock_show.called


def test_historical_first_difference_rule_missing_inflation_data(
    mock_fred_client, sample_fred_data
):
    mock_fred_client.side_effect = [
        None, sample_fred_data, sample_fred_data,
        sample_fred_data, sample_fred_data
    ]

    indicators = EconomicIndicators(
        inflation_series_id='PCETRIM12M159SFRBDAL',
        unemployment_rate_series_id='UNRATE',
        natural_unemployment_series_id='NROU'
    )

    params = FirstDifferenceRuleParameters(
        inflation_target=2.0,
        alpha=0.5,
        rho=0.5,
        elb=0.125,
        apply_elb=True
    )

    with pytest.raises(ValueError):
        historical_first_difference_rule(indicators, params)


def test_historical_first_difference_rule_missing_unemployment_data(
    mock_fred_client, sample_fred_data
):
    mock_fred_client.side_effect = [
        sample_fred_data, None, sample_fred_data,
        sample_fred_data, sample_fred_data
    ]

    indicators = EconomicIndicators(
        inflation_series_id='PCETRIM12M159SFRBDAL',
        unemployment_rate_series_id='UNRATE',
        natural_unemployment_series_id='NROU'
    )

    params = FirstDifferenceRuleParameters(
        inflation_target=2.0,
        alpha=0.5,
        rho=0.5,
        elb=0.125,
        apply_elb=True
    )

    with pytest.raises(ValueError):
        historical_first_difference_rule(indicators, params)


def test_historical_first_difference_rule_missing_natural_unemployment_data(
    mock_fred_client, sample_fred_data
):
    mock_fred_client.side_effect = [
        sample_fred_data, sample_fred_data, None,
        sample_fred_data, sample_fred_data
    ]

    indicators = EconomicIndicators(
        inflation_series_id='PCETRIM12M159SFRBDAL',
        unemployment_rate_series_id='UNRATE',
        natural_unemployment_series_id='NROU'
    )

    params = FirstDifferenceRuleParameters(
        inflation_target=2.0,
        alpha=0.5,
        rho=0.5,
        elb=0.125,
        apply_elb=True
    )

    with pytest.raises(ValueError):
        historical_first_difference_rule(indicators, params)


def test_historical_first_difference_rule_missing_fed_rate_data(
    mock_fred_client, sample_fred_data
):
    mock_fred_client.side_effect = [
        sample_fred_data, sample_fred_data, sample_fred_data,
        sample_fred_data, None
    ]

    indicators = EconomicIndicators(
        inflation_series_id='PCETRIM12M159SFRBDAL',
        unemployment_rate_series_id='UNRATE',
        natural_unemployment_series_id='NROU'
    )

    params = FirstDifferenceRuleParameters(
        inflation_target=2.0,
        alpha=0.5,
        rho=0.5,
        elb=0.125,
        apply_elb=True
    )

    with pytest.raises(ValueError):
        historical_first_difference_rule(indicators, params)


if __name__ == '__main__':
    pytest.main()



# tests/test_taylor_rule.py

from unittest.mock import patch
import pandas as pd
import pytest

from pyeconomics.data.economic_indicators import EconomicIndicators
from pyeconomics.data.model_parameters import TaylorRuleParameters

from pyeconomics.models.monetary_policy.taylor_rule import (
    taylor_rule, historical_taylor_rule, plot_historical_taylor_rule
)


@pytest.fixture
def mock_fred_data():
    """Fixture for mock FRED data."""
    return {
        'inflation_rate': 2.5,
        'unemployment_rate': 4.0,
        'natural_unemployment_rate': 4.5,
        'real_interest_rate': 1.0,
        'current_fed_rate': 0.5
    }


@patch('pyeconomics.models.monetary_policy.taylor_rule.fred_client')
def test_taylor_rule(mock_fred_client, mock_fred_data):
    mock_fred_client.get_data_or_fetch.side_effect = (
        lambda default, series_id, periods=0: mock_fred_data.get(series_id)
    )
    mock_fred_client.get_latest_value.return_value = (
        mock_fred_data['current_fed_rate']
    )

    indicators = EconomicIndicators(
        inflation_series_id='inflation_rate',
        unemployment_rate_series_id='unemployment_rate',
        natural_unemployment_series_id='natural_unemployment_rate',
        real_interest_rate_series_id='real_interest_rate'
    )

    params = TaylorRuleParameters(
        inflation_target=2.0, alpha=0.5, beta=0.5, rho=0.0, apply_elb=False
    )

    # Default test case
    result = taylor_rule(indicators, params)
    expected_result = 4.25  # Computed based on the mock data and parameters
    assert result == expected_result

    # Test with apply_elb = True
    params.apply_elb = True
    result = taylor_rule(indicators, params)
    expected_result = max(4.25, 0.125)  # Adjusted expected value with ELB
    assert result == expected_result

    # Test with verbose output
    with patch('pyeconomics.models.monetary_policy.taylor_rule.'
               'verbose_taylor_rule') as mock_verbose:
        params.verbose = True
        taylor_rule(indicators, params)
        assert mock_verbose.called

    # Test ValueError raise condition
    with pytest.raises(ValueError):
        invalid_indicators = EconomicIndicators(
            inflation_series_id='missing_series_id',
            unemployment_rate_series_id='missing_series_id',
            natural_unemployment_series_id='missing_series_id',
            real_interest_rate_series_id='missing_series_id'
        )
        taylor_rule(invalid_indicators, params)

    # Test with current_fed_rate as None
    mock_fred_client.get_latest_value.return_value = 1.0
    indicators.current_fed_rate = None
    result = taylor_rule(indicators, params)
    expected_result = 4.25  # Computed based on the mock data and parameters
    assert result == expected_result

    # Test with current_fed_rate provided
    indicators.current_fed_rate = 2.0
    result = taylor_rule(indicators, params)
    expected_result = 4.25  # Computed based on the mock data and parameters
    assert result == expected_result


@patch('pyeconomics.models.monetary_policy.taylor_rule.fred_client')
@patch('pyeconomics.models.monetary_policy.'
       'taylor_rule.fetch_historical_fed_funds_rate')
def test_historical_taylor_rule(
        mock_fetch_historical_fed_funds_rate, mock_fred_client, mock_fred_data
):
    mock_fred_client.fetch_data.side_effect = (
        lambda series_id: pd.Series([mock_fred_data[series_id]] * 10)
    )
    mock_fetch_historical_fed_funds_rate.return_value = (
        pd.Series([mock_fred_data['current_fed_rate']] * 10)
    )

    indicators = EconomicIndicators(
        inflation_series_id='inflation_rate',
        unemployment_rate_series_id='unemployment_rate',
        natural_unemployment_series_id='natural_unemployment_rate',
        real_interest_rate_series_id='real_interest_rate'
    )

    params = TaylorRuleParameters(
        inflation_target=2.0, alpha=0.5, beta=0.5, rho=0.0, apply_elb=False
    )

    # Default test case
    result = historical_taylor_rule(indicators, params)

    # Verify the shape and content of the result DataFrame
    assert not result.empty
    # Based on mock data and parameters
    assert all(result['TaylorRule'] == 4.25)

    # Test with apply_elb = True
    params.apply_elb = True
    result = historical_taylor_rule(indicators, params)
    assert not result.empty
    assert all(result['AdjustedTaylorRule'] == max(4.25, 0.125))


@patch('pyeconomics.models.monetary_policy.taylor_rule.fred_client')
def test_historical_taylor_rule_missing_data(
        mock_fred_client, mock_fred_data
):
    mock_fred_client.side_effect = [
        mock_fred_data, mock_fred_data, mock_fred_data, mock_fred_data,
        pd.Series()
    ]

    indicators = EconomicIndicators(
        inflation_series_id='inflation_rate',
        unemployment_rate_series_id='unemployment_rate',
        natural_unemployment_series_id='natural_unemployment_rate',
        real_interest_rate_series_id='real_interest_rate'
    )

    params = TaylorRuleParameters(
        inflation_target=2.0, alpha=0.5, beta=0.5, rho=0.0, apply_elb=False
    )

    with pytest.raises(ValueError):
        historical_taylor_rule(indicators, params)


@patch('pyeconomics.models.monetary_policy.taylor_rule.plt.show')
def test_plot_historical_taylor_rule(mock_show):
    historical_rule_estimates = pd.DataFrame({
        'TaylorRule': [2.5, 2.6],
        'AdjustedTaylorRule': [2.7, 2.8],
        'FedRate': [2.0, 2.1]
    }, index=pd.to_datetime(['2020-01-01', '2020-02-01']))

    # Test plot
    plot_historical_taylor_rule(historical_rule_estimates)
    assert mock_show.called


if __name__ == '__main__':
    pytest.main()



# tests/test_tr_utils.py

import pytest
from datetime import datetime

from pyeconomics.utils.tr_utils import verbose_taylor_rule


@pytest.fixture
def mock_data():
    return {
        'current_inflation_rate': 2.5,
        'inflation_target': 2.0,
        'current_unemployment_rate': 4.0,
        'natural_unemployment_rate': 4.5,
        'long_term_real_interest_rate': 2.0,
        'current_fed_rate': 0.5,
        'inflation_gap': 0.5,
        'unemployment_gap': 0.5,
        'alpha': 0.5,
        'beta': 0.5,
        'okun_factor': 2.0,
        'unadjusted_taylor_rule': 1.0,
        'adjusted_taylor_rule_after_elb': 1.0,
        'apply_elb': False,
        'elb': 0.0,
        'rho': 0.5,
        'adjusted_taylor_rule_after_inertia': 0.75
    }


def test_verbose_taylor_rule(capsys, mock_data):
    verbose_taylor_rule(mock_data)

    captured = capsys.readouterr()
    output = captured.out

    as_of_date = datetime.now().strftime("%B %d, %Y")
    expected_output = (
        "==== Economic Indicators "
        "=================================================\n"
        "Current Inflation:                               2.50%\n"
        "Target Inflation:                                2.00%\n"
        "Current Unemployment Rate:                       4.00%\n"
        "Natural Unemployment Rate:                       4.50%\n"
        "Long-Term Real Interest Rate:                    2.00%\n"
        "Current Fed Rate:                                0.50%\n"
        f"As of Date:                                      {as_of_date}\n"
        "\n==== Gaps "
        "================================================================\n"
        "Inflation Gap:                                   0.50%\n"
        "Unemployment Gap:                                0.50%\n"
        "\n==== Taylor Rule "
        "=========================================================\n"
        "  Long-Term Real Interest Rate:                  2.00%\n"
        "  Current Inflation:                             + 2.50%\n"
        "  Alpha * Inflation Gap:                         + 0.50 * 0.50%\n"
        "  Beta * Okun Factor * Unemployment Gap:         "
        "+ 0.50 * 2.00 * 0.50%\n"
        "-------------------------------------"
        "-------------------------------------\n"
        "  Unadjusted Taylor Rule Estimate:               1.00%\n"
        "\n==== Adjusted Taylor Rule "
        "================================================\n"
        "\n  Policy Inertia Adjustment:              \n"
        "  Policy Inertia Coefficient (rho):              0.50\n"
        "  Current Fed Rate:                              * 0.50%\n"
        "  Adjustment Coefficient (1 - rho):              + (1 - 0.50)\n"
        "  Taylor Rule Adjusted for ELB:                  * 1.00%\n"
        "-------------------------------------"
        "-------------------------------------\n"
        "  Adjusted Taylor Rule Estimate:                 0.75%\n"
        "\n==== Policy Prescription "
        "=================================================\n"
        "  The Adjusted Taylor Rule Estimate is 0.25% higher than the "
        "Current \n  Fed Rate. The Fed should consider raising the interest "
        "rate by 0.25%.\n"
    )

    # Comparing the captured output with the expected output
    assert output.strip() == expected_output.strip()


def test_verbose_taylor_rule_apply_elb(capsys, mock_data):
    mock_data['apply_elb'] = True
    mock_data['adjusted_taylor_rule_after_elb'] = 1.25
    mock_data['unadjusted_taylor_rule'] = 1.0
    mock_data['elb'] = 0.5
    mock_data['adjusted_taylor_rule_after_inertia'] = 0.88
    verbose_taylor_rule(mock_data)

    captured = capsys.readouterr()
    output = captured.out

    as_of_date = datetime.now().strftime("%B %d, %Y")
    expected_output = (
        "==== Economic Indicators "
        "=================================================\n"
        "Current Inflation:                               2.50%\n"
        "Target Inflation:                                2.00%\n"
        "Current Unemployment Rate:                       4.00%\n"
        "Natural Unemployment Rate:                       4.50%\n"
        "Long-Term Real Interest Rate:                    2.00%\n"
        "Current Fed Rate:                                0.50%\n"
        f"As of Date:                                      {as_of_date}\n"
        "\n==== Gaps "
        "================================================================\n"
        "Inflation Gap:                                   0.50%\n"
        "Unemployment Gap:                                0.50%\n"
        "\n==== Taylor Rule "
        "=========================================================\n"
        "  Long-Term Real Interest Rate:                  2.00%\n"
        "  Current Inflation:                             + 2.50%\n"
        "  Alpha * Inflation Gap:                         + 0.50 * 0.50%\n"
        "  Beta * Okun Factor * Unemployment Gap:         "
        "+ 0.50 * 2.00 * 0.50%\n"
        "-------------------------------------"
        "-------------------------------------\n"
        "  Unadjusted Taylor Rule Estimate:               1.00%\n"
        "\n==== Adjusted Taylor Rule "
        "================================================\n"
        "  Effective Lower Bound (ELB) Adjustment: \n"
        "  Maximum of Taylor Rule or ELB:                 max(1.00%, 0.50%)\n"
        "-------------------------------------"
        "-------------------------------------\n"
        "  Taylor Rule Adjusted for ELB:                  1.25%\n"
        "\n  Policy Inertia Adjustment:              \n"
        "  Policy Inertia Coefficient (rho):              0.50\n"
        "  Current Fed Rate:                              * 0.50%\n"
        "  Adjustment Coefficient (1 - rho):              + (1 - 0.50)\n"
        "  Taylor Rule Adjusted for ELB:                  * 1.25%\n"
        "-------------------------------------"
        "-------------------------------------\n"
        "  Adjusted Taylor Rule Estimate:                 0.88%\n"
        "\n==== Policy Prescription "
        "=================================================\n"
        "  The Adjusted Taylor Rule Estimate is 0.38% higher than the "
        "Current \n  Fed Rate. The Fed should consider raising the interest "
        "rate by 0.50%.\n"
    )

    # Comparing the captured output with the expected output
    assert output.strip() == expected_output.strip()


def test_verbose_taylor_rule_lower_rate(capsys, mock_data):
    mock_data['adjusted_taylor_rule_after_inertia'] = 0.25
    verbose_taylor_rule(mock_data)

    captured = capsys.readouterr()
    output = captured.out

    as_of_date = datetime.now().strftime("%B %d, %Y")
    expected_output = (
        "==== Economic Indicators "
        "=================================================\n"
        "Current Inflation:                               2.50%\n"
        "Target Inflation:                                2.00%\n"
        "Current Unemployment Rate:                       4.00%\n"
        "Natural Unemployment Rate:                       4.50%\n"
        "Long-Term Real Interest Rate:                    2.00%\n"
        "Current Fed Rate:                                0.50%\n"
        f"As of Date:                                      {as_of_date}\n"
        "\n==== Gaps "
        "================================================================\n"
        "Inflation Gap:                                   0.50%\n"
        "Unemployment Gap:                                0.50%\n"
        "\n==== Taylor Rule "
        "=========================================================\n"
        "  Long-Term Real Interest Rate:                  2.00%\n"
        "  Current Inflation:                             + 2.50%\n"
        "  Alpha * Inflation Gap:                         + 0.50 * 0.50%\n"
        "  Beta * Okun Factor * Unemployment Gap:         "
        "+ 0.50 * 2.00 * 0.50%\n"
        "-------------------------------------"
        "-------------------------------------\n"
        "  Unadjusted Taylor Rule Estimate:               1.00%\n"
        "\n==== Adjusted Taylor Rule "
        "================================================\n"
        "\n  Policy Inertia Adjustment:              \n"
        "  Policy Inertia Coefficient (rho):              0.50\n"
        "  Current Fed Rate:                              * 0.50%\n"
        "  Adjustment Coefficient (1 - rho):              + (1 - 0.50)\n"
        "  Taylor Rule Adjusted for ELB:                  * 1.00%\n"
        "-------------------------------------"
        "-------------------------------------\n"
        "  Adjusted Taylor Rule Estimate:                 0.25%\n"
        "\n==== Policy Prescription "
        "=================================================\n"
        "  The Adjusted Taylor Rule Estimate is 0.25% lower than the "
        "Current \n  Fed Rate. The Fed should consider lowering the interest "
        "rate by -0.25%.\n"
    )

    # Comparing the captured output with the expected output
    assert output.strip() == expected_output.strip()


def test_verbose_taylor_rule_maintain_rate(capsys, mock_data):
    mock_data['adjusted_taylor_rule_after_inertia'] = 0.5
    verbose_taylor_rule(mock_data)

    captured = capsys.readouterr()
    output = captured.out

    as_of_date = datetime.now().strftime("%B %d, %Y")
    expected_output = (
        "==== Economic Indicators "
        "=================================================\n"
        "Current Inflation:                               2.50%\n"
        "Target Inflation:                                2.00%\n"
        "Current Unemployment Rate:                       4.00%\n"
        "Natural Unemployment Rate:                       4.50%\n"
        "Long-Term Real Interest Rate:                    2.00%\n"
        "Current Fed Rate:                                0.50%\n"
        f"As of Date:                                      {as_of_date}\n"
        "\n==== Gaps "
        "================================================================\n"
        "Inflation Gap:                                   0.50%\n"
        "Unemployment Gap:                                0.50%\n"
        "\n==== Taylor Rule "
        "=========================================================\n"
        "  Long-Term Real Interest Rate:                  2.00%\n"
        "  Current Inflation:                             + 2.50%\n"
        "  Alpha * Inflation Gap:                         + 0.50 * 0.50%\n"
        "  Beta * Okun Factor * Unemployment Gap:         "
        "+ 0.50 * 2.00 * 0.50%\n"
        "-------------------------------------"
        "-------------------------------------\n"
        "  Unadjusted Taylor Rule Estimate:               1.00%\n"
        "\n==== Adjusted Taylor Rule "
        "================================================\n"
        "\n  Policy Inertia Adjustment:              \n"
        "  Policy Inertia Coefficient (rho):              0.50\n"
        "  Current Fed Rate:                              * 0.50%\n"
        "  Adjustment Coefficient (1 - rho):              + (1 - 0.50)\n"
        "  Taylor Rule Adjusted for ELB:                  * 1.00%\n"
        "-------------------------------------"
        "-------------------------------------\n"
        "  Adjusted Taylor Rule Estimate:                 0.50%\n"
        "\n==== Policy Prescription "
        "=================================================\n"
        "  The Adjusted Taylor Rule Estimate is equal to the Current "
        "Fed Rate.\n  The Fed should maintain the current interest rate.\n"
    )

    # Comparing the captured output with the expected output
    assert output.strip() == expected_output.strip()


if __name__ == '__main__':
    pytest.main()



# tests/test_fred_api.py

import os
import datetime
import pytest
from unittest.mock import patch, MagicMock

import pandas as pd

from pyeconomics.api.fred_api import FredClient


@pytest.fixture(scope='module')
def fred_client():
    FredClient.reset_instance()
    api_key = 'test_api_key'
    return FredClient(api_key)


def test_singleton_pattern(fred_client):
    """
    Test to ensure FredClient follows the singleton pattern.
    """
    with patch.object(FredClient, '__new__', return_value=fred_client):
        another_instance = FredClient()
        assert fred_client is another_instance


def test_singleton_none_instance():
    FredClient.reset_instance()
    assert FredClient._instance is None


def test_singleton_existing_instance(fred_client):
    """
    Test to ensure FredClient does not create a new instance if one already
    exists.
    """
    FredClient.reset_instance()  # Ensure a fresh start
    first_instance = FredClient(api_key='test_api_key')
    second_instance = FredClient(api_key='another_test_api_key')

    assert first_instance is second_instance  # Should be the same instance


def test_fetch_data_from_cache(fred_client):
    with patch('pyeconomics.api.fred_api.load_from_cache') as mock_load_cache:
        series_id = 'GDP'
        expected_data = pd.Series([1, 2, 3], name=series_id)
        mock_load_cache.return_value = expected_data

        data = fred_client.fetch_data(series_id)
        assert data.equals(expected_data)
        mock_load_cache.assert_called_once_with(f'fred_series_{series_id}')


def test_fetch_data_from_api(fred_client):
    with patch('pyeconomics.api.fred_api.load_from_cache') as mock_load_cache, \
        patch('pyeconomics.api.fred_api.save_to_cache') as mock_save_cache, \
            patch.object(fred_client.client, 'get_series') as mock_get_series:
        mock_load_cache.return_value = None
        series_id = 'GDP'
        expected_data = pd.Series([1, 2, 3], name=series_id)
        mock_get_series.return_value = expected_data

        data = fred_client.fetch_data(series_id)
        assert data.equals(expected_data)
        mock_get_series.assert_called_once_with(series_id)
        mock_save_cache.assert_called_once_with(f'fred_series_{series_id}',
                                                expected_data)


def test_get_latest_value(fred_client):
    with patch.object(fred_client, 'fetch_data') as mock_fetch_data:
        series_id = 'GDP'
        today = datetime.date.today()
        data = pd.Series(
            [1, 2, 3],
            index=pd.date_range(end=today, periods=3, freq='D'),
            name=series_id
        )
        mock_fetch_data.return_value = data

        latest_value = fred_client.get_latest_value(series_id)
        assert latest_value == 3


def test_get_historical_value(fred_client):
    with patch.object(fred_client, 'fetch_data') as mock_fetch_data:
        series_id = 'GDP'
        data = pd.Series([1, 2, 3], name=series_id)
        mock_fetch_data.return_value = data

        historical_value = fred_client.get_historical_value(
            series_id, periods=-2)
        assert historical_value == 2


def test_get_series_name(fred_client):
    with patch.object(
            fred_client.client, 'get_series_info'
    ) as mock_get_series_info:
        series_id = 'GDP'
        series_info = {"title": "Gross Domestic Product"}
        mock_get_series_info.return_value = series_info

        series_name = fred_client.get_series_name(series_id)
        assert series_name == "Gross Domestic Product"


def test_get_data_or_fetch(fred_client):
    FredClient.reset_instance()  # Reset instance before test
    FredClient._instance = fred_client  # Ensure the instance is set
    with (patch.object(
        FredClient, 'get_latest_value'
    ) as mock_get_latest_value, patch.object(
        FredClient, 'get_historical_value'
    ) as mock_get_historical_value):
        series_id = 'GDP'
        default = None
        periods = 0
        expected_value = 3

        mock_get_latest_value.return_value = expected_value
        value = FredClient.get_data_or_fetch(default, series_id, periods)
        assert value == expected_value

        periods = -2
        mock_get_historical_value.return_value = 2
        value = FredClient.get_data_or_fetch(default, series_id, periods)
        assert value == 2


def test_keyring_unavailable():
    with patch.dict('sys.modules', {'keyring': None}):
        from importlib import reload
        FredClient.reset_instance()

        # Set a test API key in the environment before reloading
        with patch.dict(
            'os.environ',
            {'FRED_API_KEY': 'test_api_key'}
        ):
            import pyeconomics.api.fred_api as fred_api
            fred_api.KEYRING_AVAILABLE = False
            reload(fred_api)

            assert not fred_api.KEYRING_AVAILABLE

            # Use a context manager to create the FredClient instance
            with patch(
                'pyeconomics.api.fred_api.FredClient.__new__',
                    return_value=MagicMock()):
                # Now the FRED_API_KEY environment variable will be used
                client = FredClient(api_key=None)
                assert client is not None


def test_keyring_available_but_no_api_key():
    with patch('keyring.get_password', return_value=None):
        FredClient.reset_instance()  # Ensure fresh instance for each test
        # Ensure no environment variable is set
        with patch.dict('os.environ', {}, clear=True):
            with pytest.raises(
                ValueError,
                match="API Key for FRED must be provided "
                      "or retrievable from keyring."
            ):
                FredClient(api_key=None)  # No API key provided


def test_no_api_key_provided():
    with patch('keyring.get_password', return_value=None):
        FredClient.reset_instance()  # Ensure fresh instance for each test
        # Ensure no environment variable is set
        with patch.dict('os.environ', {}, clear=True):
            with pytest.raises(
                ValueError,
                match="API Key for FRED must be provided or "
                      "retrievable from keyring."
            ):
                FredClient(api_key=None)  # No API key provided


def test_fetch_data_no_data_found(fred_client):
    with patch.object(
        fred_client.client,
        'get_series',
            return_value=pd.Series([])):
        with pytest.raises(ValueError, match="No data found for series ID GDP"):
            fred_client.fetch_data('GDP')


def test_fetch_data_exception_handling(fred_client):
    with patch.object(
        fred_client.client,
        'get_series',
            side_effect=Exception("API error")):
        with pytest.raises(Exception, match="API error"):
            fred_client.fetch_data('GDP')


def test_get_latest_value_exception_handling(fred_client):
    with patch.object(
        fred_client,
        'fetch_data',
            side_effect=Exception("Data fetch error")):
        with pytest.raises(Exception, match="Data fetch error"):
            fred_client.get_latest_value('GDP')


def test_get_historical_value_exception_handling(fred_client):
    with patch.object(
        fred_client,
        'fetch_data',
            side_effect=Exception("Data fetch error")
    ):
        with pytest.raises(Exception, match="Data fetch error"):
            fred_client.get_historical_value('GDP', periods=-2)


def test_get_series_name_exception_handling(fred_client):
    with patch.object(
        fred_client.client,
        'get_series_info',
            side_effect=Exception("API error")):
        with pytest.raises(Exception, match="API error"):
            fred_client.get_series_name('GDP')


def test_keyring_get_password_called():
    with patch.dict('sys.modules', {'keyring': MagicMock()}):
        from importlib import reload
        import pyeconomics.api.fred_api as fred_api
        reload(fred_api)
        FredClient.reset_instance()

        mock_keyring = fred_api.keyring
        mock_keyring.get_password = MagicMock(return_value='mocked_api_key')

        # Ensure that KEYRING_AVAILABLE is set to True
        with patch('pyeconomics.api.fred_api.KEYRING_AVAILABLE', True):
            with patch.dict(os.environ, {'FRED_API_KEY': ''}):
                try:
                    FredClient(api_key=None)
                except ValueError:
                    # If the initialization fails,
                    # catch the exception to check why
                    pass

                mock_keyring.get_password.assert_called_once_with(
                    'fred', 'api_key')


def test_get_data_or_fetch_exception_handling():
    FredClient.reset_instance()
    with patch.object(
        FredClient, 'get_latest_value', side_effect=Exception("fetch error")
    ):
        with patch('pyeconomics.api.fred_api.KEYRING_AVAILABLE', False):
            FredClient._instance = FredClient(api_key="test_api_key")
            with pytest.raises(Exception, match="fetch error"):
                FredClient.get_data_or_fetch(None, 'GDP', 0)


def test_instance_creation_with_existing_instance():
    FredClient.reset_instance()
    with patch.object(
            FredClient, '__new__', wraps=FredClient.__new__) as mock_new:
        first_instance = FredClient(api_key='test_api_key')
        second_instance = FredClient(api_key='another_test_api_key')

        assert first_instance is second_instance  # Should be the same instance
        # Ensure __new__ was called twice, but instance creation only once
        assert mock_new.call_count == 2
        assert first_instance == second_instance


if __name__ == '__main__':
    pytest.main()



# tests/test_balanced_approach_rule.py

import os
import shutil
from datetime import datetime, timedelta
from hashlib import sha256

import pytest

from pyeconomics.api.cache_manager import (
    CACHE_DIR, cache_filename, save_to_cache, load_from_cache
)


@pytest.fixture(scope='function', autouse=True)
def setup_and_teardown():
    """Setup and teardown for tests."""
    # Ensure the cache directory exists before each test
    if not os.path.exists(CACHE_DIR):
        os.makedirs(CACHE_DIR)
    yield
    # Cleanup after test
    if os.path.exists(CACHE_DIR):
        shutil.rmtree(CACHE_DIR)


def test_cache_filename():
    key = 'test_key'
    filename = cache_filename(key)
    hashed_key = sha256(key.encode('utf-8')).hexdigest()
    expected_filename = os.path.join(CACHE_DIR, f"{hashed_key}.pkl")
    assert filename == expected_filename


def test_save_to_cache():
    key = 'test_key'
    data = {'value': 42}
    save_to_cache(key, data)
    filename = cache_filename(key)
    assert os.path.exists(filename)


def test_load_from_cache():
    key = 'test_key'
    data = {'value': 42}
    save_to_cache(key, data)
    loaded_data = load_from_cache(key)
    assert loaded_data == data


def test_load_from_cache_expired():
    key = 'test_key'
    data = {'value': 42}
    save_to_cache(key, data)
    # Manually set the file's modified time to simulate expiry
    filename = cache_filename(key)
    old_time = datetime.now() - timedelta(days=2)
    old_timestamp = old_time.timestamp()
    os.utime(filename, (old_timestamp, old_timestamp))
    loaded_data = load_from_cache(key, expiry=timedelta(days=1))
    assert loaded_data is None


def test_load_from_cache_nonexistent_key():
    key = 'nonexistent_key'
    loaded_data = load_from_cache(key)
    assert loaded_data is None


def test_cache_directory_creation():
    # Ensure the directory exists before running the test
    if not os.path.exists(CACHE_DIR):
        os.makedirs(CACHE_DIR)
    assert os.path.exists(CACHE_DIR)


def test_cache_directory_creation_when_not_exists():
    """Test creation of cache directory when it does not exist."""
    # Remove the cache directory if it exists
    if os.path.exists(CACHE_DIR):
        shutil.rmtree(CACHE_DIR)
    assert not os.path.exists(CACHE_DIR)  # Ensure the directory does not exist
    # Re-import cache_manager to trigger the directory creation
    import importlib
    import pyeconomics.api.cache_manager as cm
    importlib.reload(cm)
    assert os.path.exists(CACHE_DIR)  # Now it should exist


if __name__ == '__main__':
    pytest.main()
